/* config_callbacks.c -- callbacks for the configure window
 *
 * Copyright (C) 2001 Janusz Gregorczyk <jgregor@kki.net.pl>
 *
 * This file is part of xvs.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include <pthread.h>
#include <string.h>

#include "config_callbacks.h"
#include "config_interface.h"
#include "support.h"
#include "../preset.h"

/* Externs. */

extern GtkWidget *configure_window;
GtkWidget *preset_tree, *option_frame, *option_scrolled_window;
GtkWidget *button_add, *button_remove, *button_clone;
GtkCTreeNode *selected_preset_node;

extern XVS_Effect *root_fx;
extern pthread_mutex_t root_fx_mutex;
extern gchar *XVS_get_preset_dir (void);

/* Functions. */


static void option_window_remove_all_cb (GtkWidget *widget, gpointer data) {
  gtk_container_remove (GTK_CONTAINER(option_scrolled_window), widget);
}


static void add_fx (XVS_Effect *fx, GtkCTreeNode *parent, GtkCTreeNode *sibling,
		    gboolean initonly) {
  GtkCTreeNode *node;
  gchar *desc = fx->desc->get_desc ();

  node = gtk_ctree_insert_node (GTK_CTREE(preset_tree), parent, sibling, &desc,
				0, NULL, NULL, NULL, NULL, 
				fx->desc->flags & XVS_FLAG_CONTAINER
				? FALSE : TRUE, TRUE);

  /* Don't connect modules if we are only building a tree. */
  if (!initonly) {
    if (parent) {			/* Add fx to a container. */
      XVS_Effect *fx_sibling = NULL;
      XVS_Effect *container = (XVS_Effect *)gtk_ctree_node_get_row_data 
	(GTK_CTREE (preset_tree), parent);
      if (sibling)
	fx_sibling = (XVS_Effect *)gtk_ctree_node_get_row_data 
	  (GTK_CTREE (preset_tree), sibling);
      pthread_mutex_lock (&root_fx_mutex);
      container_add_fx_before (container, fx, fx_sibling);
      pthread_mutex_unlock (&root_fx_mutex);
    } else {			/* There's no container -- make fx root. */
      pthread_mutex_lock (&root_fx_mutex);
      g_assert (root_fx == NULL);
      root_fx = fx;
      pthread_mutex_unlock (&root_fx_mutex);
    }
  }

  gtk_ctree_node_set_row_data (GTK_CTREE (preset_tree), node, fx);

  /* Add children of the container. */
  if (fx->desc->flags & XVS_FLAG_CONTAINER) {
    GList *child;
    child = fx->children;
    while (child) {
      add_fx ((XVS_Effect *)child->data, node, NULL, TRUE);
      child = child->next;
    }
  }

  if (!initonly) 
    gtk_ctree_select (GTK_CTREE(preset_tree), node);
}


static void remove_node (GtkCTreeNode *node) {
  XVS_Effect *fx, *parent_fx;
  GtkCTreeNode *parent;

  fx = (XVS_Effect *)gtk_ctree_node_get_row_data (GTK_CTREE (preset_tree), node);

  parent = GTK_CTREE_ROW(node)->parent;

  if (parent) {
    pthread_mutex_lock (&root_fx_mutex);
    parent_fx = (XVS_Effect *)gtk_ctree_node_get_row_data 
      (GTK_CTREE (preset_tree), parent);
    gtk_ctree_remove_node (GTK_CTREE(preset_tree), node);
    container_remove_fx (parent_fx, fx);
    pthread_mutex_unlock (&root_fx_mutex);
  } else {			/* It's the root node. */
    pthread_mutex_lock (&root_fx_mutex);
    gtk_ctree_remove_node (GTK_CTREE(preset_tree), node);
    g_assert (root_fx == fx);
    root_fx = NULL;
    pthread_mutex_unlock (&root_fx_mutex);
  }

  fx_destroy (fx);

  if (!root_fx)
    gtk_widget_set_sensitive (button_add, TRUE);
}


void rebuild_preset_tree (void) {
  GtkCTreeNode *root;

  /* Kill the tree. */
  root = gtk_ctree_node_nth (GTK_CTREE(preset_tree), 0);
  if (root)
    gtk_ctree_remove_node (GTK_CTREE(preset_tree), root);

  /* Rebuild the tree. */
  if (root_fx)
    add_fx (root_fx, NULL, NULL, TRUE);

  gtk_ctree_expand_recursive (GTK_CTREE(preset_tree), NULL);

  selected_preset_node = NULL;
  gtk_widget_set_sensitive (button_add, root_fx ? FALSE : TRUE);
  gtk_widget_set_sensitive (button_remove, FALSE);
  gtk_widget_set_sensitive (button_clone, FALSE);
}


static void file_open_ok (GtkWidget *widget, GtkFileSelection *fs) {
  pthread_mutex_lock (&root_fx_mutex);
  {
    /* Kill the current preset. */
    if (root_fx)
      fx_destroy (root_fx);
    
    root_fx = preset_load (gtk_file_selection_get_filename 
			   (GTK_FILE_SELECTION(fs)));
  }
  pthread_mutex_unlock (&root_fx_mutex);

  rebuild_preset_tree ();

  /* Kill the dialog. */
  gtk_widget_destroy (GTK_WIDGET(fs));
}


static void file_save_as_ok (GtkWidget *widget, GtkFileSelection *fs) {
  pthread_mutex_lock (&root_fx_mutex);
  {
    g_assert (root_fx);
    
    preset_save (root_fx, gtk_file_selection_get_filename 
		 (GTK_FILE_SELECTION(fs)));
  }
  pthread_mutex_unlock (&root_fx_mutex);

  /* Kill the dialog. */
  gtk_widget_destroy (GTK_WIDGET(fs));
}


static void open_file_dialog (gboolean save) {
  GtkWidget *file_sel;

  /* Create a new file selection widget */
  file_sel = gtk_file_selection_new (save ? "Save as" : "Open");

  /* Move to default preset directory. */
  {
    gchar *dirname, *path;
    dirname = XVS_get_preset_dir ();
    path = g_strconcat (dirname, "/", NULL);
    g_free (dirname);
    gtk_file_selection_set_filename (GTK_FILE_SELECTION(file_sel), path);
    g_free (path);
  }
    
  /* Connect the ok_button to file_whatever_ok function */
  gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION(file_sel)->ok_button),
		      "clicked",  save ? 
		      (GtkSignalFunc)file_save_as_ok :
		      (GtkSignalFunc)file_open_ok,
		      file_sel);
  
  /* Connect the cancel_button to destroy the widget */
  gtk_signal_connect_object 
    (GTK_OBJECT(GTK_FILE_SELECTION(file_sel)->cancel_button),
     "clicked", (GtkSignalFunc)gtk_widget_destroy, 
     GTK_OBJECT(file_sel));
  
  gtk_widget_show(file_sel);
}


static void enabled_toggled (GtkToggleButton *button, XVS_Effect *fx) {
  if (gtk_toggle_button_get_active (button))
    fx->state |= XVS_FX_ENABLED;
  else
    fx->state &= ~XVS_FX_ENABLED;
}


static void fill_option_frame (XVS_Effect *fx) {
  GtkWidget *vbox;

  g_assert (fx);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_border_width (GTK_CONTAINER(vbox), 5);
  gtk_scrolled_window_add_with_viewport 
    (GTK_SCROLLED_WINDOW(option_scrolled_window), vbox);
  gtk_widget_show (vbox);
  
  {
    GtkWidget *button, *widget;

    button = gtk_check_button_new_with_label ("Enabled");
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), 
				  fx->state & XVS_FX_ENABLED);
    gtk_signal_connect (GTK_OBJECT(button), "toggled", 
			GTK_SIGNAL_FUNC(enabled_toggled), 
			fx);
    gtk_box_pack_start (GTK_BOX(vbox), button, FALSE, FALSE, 0);
    gtk_widget_show (button);
  
    gtk_frame_set_label (GTK_FRAME(option_frame), fx->desc->get_desc ());

    if (fx->desc->configure && (widget = fx->desc->configure (fx))) {
      gtk_box_pack_start (GTK_BOX(vbox), widget, FALSE, FALSE, 0);
      gtk_widget_show (widget);
    }
  }
}


static void menuitem_response (XVS_Effect_Desc *desc) {
  XVS_Effect *fx, *node_fx;
  GtkCTreeNode *parent, *sibling;
 
  /* Create a new preset entry. */
  fx = fx_create (desc);

  if (selected_preset_node) 
    node_fx = (XVS_Effect *)gtk_ctree_node_get_row_data 
      (GTK_CTREE (preset_tree), selected_preset_node);
  else
    node_fx = NULL;

  /* Determine position. */
  if (selected_preset_node) {
    if (node_fx && (node_fx->desc->flags & XVS_FLAG_CONTAINER)) {
      parent = selected_preset_node;
      sibling = NULL;
    } else {
      parent = GTK_CTREE_ROW(selected_preset_node)->parent;
      sibling = selected_preset_node;
    }
  } else {
    parent = sibling = NULL;
  }

  /* Put it on the tree. */
  add_fx (fx, parent, sibling, FALSE);
}


static GtkWidget *fx_menu_create (void) {
  const XVS_Effect_Desc **fx = &fxdrivers[0];
  GtkWidget *fx_menu, *filter_menu, *trans_menu, *render_menu, *container_menu;
  GtkWidget *menuitem;

  /* Create a popup menu. */
  fx_menu = gtk_menu_new ();

  /* Create submenus. */
  menuitem = gtk_menu_item_new_with_label ("Filter");
  gtk_menu_append (GTK_MENU (fx_menu), menuitem);
  filter_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM(menuitem), filter_menu);
  gtk_widget_show (filter_menu);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new_with_label ("Trans");
  gtk_menu_append (GTK_MENU (fx_menu), menuitem);
  trans_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM(menuitem), trans_menu);
  gtk_widget_show (trans_menu);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new_with_label ("Render");
  gtk_menu_append (GTK_MENU (fx_menu), menuitem);
  render_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM(menuitem), render_menu);
  gtk_widget_show (render_menu);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new_with_label ("Container");
  gtk_menu_append (GTK_MENU (fx_menu), menuitem);
  container_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM(menuitem), container_menu);
  gtk_widget_show (container_menu);
  gtk_widget_show (menuitem);

  /* */

  while (*fx) {
    /* Create a new menu item. */
    menuitem = gtk_menu_item_new_with_label ((*fx)->get_desc ());

    if (!strncmp ((*fx)->id, "Filter", 6))
      gtk_menu_append (GTK_MENU (filter_menu), menuitem);
    else if (!strncmp ((*fx)->id, "Trans", 5))
      gtk_menu_append (GTK_MENU (trans_menu), menuitem);
    else if (!strncmp ((*fx)->id, "Render", 6))
      gtk_menu_append (GTK_MENU (render_menu), menuitem);
    else if (!strncmp ((*fx)->id, "Container", 9))
      gtk_menu_append (GTK_MENU (container_menu), menuitem);
    else
      gtk_menu_append (GTK_MENU (fx_menu), menuitem);

    gtk_signal_connect_object (GTK_OBJECT (menuitem), "activate",
       GTK_SIGNAL_FUNC (menuitem_response), (gpointer)*fx);
    gtk_widget_show (menuitem);

    fx++;
  }

  return fx_menu;
}

/* Callbacks. */

gboolean
on_configure_window_delete_event       (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  preset_tree = NULL;
  option_frame = NULL;
  option_scrolled_window = NULL;

  configure_window = NULL;

  return FALSE;
}


gboolean drag_compare_func (GtkCTree *ctree, GtkCTreeNode *source_node,
			    GtkCTreeNode *new_parent, GtkCTreeNode *new_sibling) {
  /* Don't let the node to be dragged over the root node. */
  if (new_parent == NULL)
    return FALSE;

  return TRUE;
}


void
on_configure_window_show               (GtkWidget       *widget,
                                        gpointer         user_data)
{
  /* Lookup some widgets. */
  preset_tree = lookup_widget (configure_window, "preset_tree");

  button_add = lookup_widget (configure_window, "button_add");
  button_remove = lookup_widget (configure_window, "button_remove");
  button_clone = lookup_widget (configure_window, "button_clone");

  option_scrolled_window = lookup_widget (configure_window, 
					  "option_scrolled_window");
  option_frame = lookup_widget (configure_window, "option_frame");
  

  gtk_ctree_set_reorderable (GTK_CTREE(preset_tree), TRUE);
  gtk_ctree_set_drag_compare_func (GTK_CTREE(preset_tree), 
				   (GtkCTreeCompareDragFunc)drag_compare_func);

  /* Fill the tree. */
  rebuild_preset_tree ();
}


void
on_file_new_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkCTreeNode *root;

  root = gtk_ctree_node_nth (GTK_CTREE(preset_tree), 0);
  if (root)
    gtk_ctree_remove_node (GTK_CTREE(preset_tree), root);

  pthread_mutex_lock (&root_fx_mutex);
  {
    if (root_fx) {
      fx_destroy (root_fx);
      root_fx = NULL;
    }
    
    /* There's no root -- we might add one but we cannot remove anything. */
    gtk_widget_set_sensitive (button_add, TRUE);
    gtk_widget_set_sensitive (button_remove, FALSE);
    gtk_widget_set_sensitive (button_clone, FALSE);
  }
  pthread_mutex_unlock (&root_fx_mutex);
}


void
on_file_open_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  open_file_dialog (FALSE);
}


void
on_file_save_as_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  open_file_dialog (TRUE);
}


void
on_file_close_activate                 (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gtk_widget_destroy (configure_window);
  configure_window = NULL;
}


void
on_button_add_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *menu = fx_menu_create ();
  gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL, 0, 0);
}


void
on_button_remove_clicked               (GtkButton       *button,
                                        gpointer         user_data)
{
  remove_node (selected_preset_node);
}


void
on_button_clone_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
  XVS_Effect *fx, *copy;
  GtkCTreeNode *parent, *sibling;

  fx = (XVS_Effect *)gtk_ctree_node_get_row_data
    (GTK_CTREE(preset_tree), selected_preset_node);

  copy = fx_clone (fx);

  parent = GTK_CTREE_ROW(selected_preset_node)->parent;
  sibling = selected_preset_node;

  add_fx (copy, parent, sibling, FALSE);
}


void
on_preset_tree_select_row         (GtkCTree        *ctree,
				   GList           *node,
				   gint             column,
				   gpointer         user_data)
{
  XVS_Effect *fx;
  GtkCTreeNode *selection = (GtkCTreeNode *)node;

  fx = (XVS_Effect *)gtk_ctree_node_get_row_data (ctree, selection);

  if (fx)
    fill_option_frame (fx);

  gtk_widget_set_sensitive (button_remove, TRUE);
  gtk_widget_set_sensitive (button_clone,
			    GTK_CTREE_ROW(node)->parent ? TRUE : FALSE);

  if (fx->desc->flags & XVS_FLAG_CONTAINER) {
    gtk_widget_set_sensitive (button_add, TRUE);
  } else {
    gtk_widget_set_sensitive (button_add,
			      GTK_CTREE_ROW(node)->parent ? TRUE : FALSE);
  }

  selected_preset_node = selection;
}


void
on_preset_tree_unselect_row       (GtkCTree        *ctree,
				   GList           *node,
				   gint             column,
				   gpointer         user_data)
{
  gtk_frame_set_label (GTK_FRAME(option_frame), NULL);

  gtk_container_foreach (GTK_CONTAINER(option_scrolled_window), 
			 option_window_remove_all_cb, NULL);

  gtk_widget_set_sensitive (button_add, FALSE);
  gtk_widget_set_sensitive (button_remove, FALSE);
  gtk_widget_set_sensitive (button_clone, FALSE);
  selected_preset_node = NULL;
}


void
on_preset_tree_move                    (GtkCTree        *ctree,
                                        GList           *node,
                                        GList           *new_parent,
                                        GList           *new_sibling,
                                        gpointer         user_data)
{
  XVS_Effect *fx, *parent_fx, *new_parent_fx = NULL, *new_sibling_fx = NULL;
  GtkCTreeNode *parent;

  pthread_mutex_lock (&root_fx_mutex);
  {
    fx = (XVS_Effect *)gtk_ctree_node_get_row_data 
      (GTK_CTREE (preset_tree), (GtkCTreeNode*)node);
    
    g_assert (fx);
    
    /* Find a container to which `node' belongs. */
    if ((parent = GTK_CTREE_ROW(node)->parent)) {
      parent_fx = (XVS_Effect *)gtk_ctree_node_get_row_data 
	(GTK_CTREE (preset_tree), parent);
    }
    
    if (new_parent)
      new_parent_fx = (XVS_Effect *)gtk_ctree_node_get_row_data 
	(GTK_CTREE (preset_tree), (GtkCTreeNode*)new_parent);
    
    if (new_sibling)
      new_sibling_fx = (XVS_Effect *)gtk_ctree_node_get_row_data 
	(GTK_CTREE (preset_tree), (GtkCTreeNode*)new_sibling);
    
    g_assert (parent_fx);
    g_assert (new_parent_fx);
    g_assert (parent_fx->desc->flags & XVS_FLAG_CONTAINER);
    g_assert (new_parent_fx->desc->flags & XVS_FLAG_CONTAINER);
    
    /* Remove... */
    container_remove_fx (parent_fx, fx);
    
    /* ...and move. */
    container_add_fx_before (new_parent_fx, fx, new_sibling_fx);
  }

  pthread_mutex_unlock (&root_fx_mutex);
}
