/* popup_callbacks.c -- callbacks for the popup menu
 *
 * Copyright (C) 2001 Janusz Gregorczyk <jgregor@kki.net.pl>
 *
 * This file is part of xvs.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <dirent.h>
#include <gtk/gtk.h>
#include <pthread.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "popup_callbacks.h"
#include "popup_interface.h"
#include "support.h"
#include "../preset.h"

/* Externs. */

extern XVS_Effect *root_fx;
extern pthread_mutex_t root_fx_mutex;
extern GtkWidget *configure_window;

extern void XVS_about (void);
extern void XVS_configure (void);
extern gchar *XVS_get_preset_dir (void);
extern gchar *XVS_get_global_preset_dir (void);
extern void rebuild_preset_tree (void);

/* Functions. */

static void on_popup_menu_item_activate (GtkMenuItem *menuitem, 
					 gpointer user_data) {
  pthread_mutex_lock (&root_fx_mutex);
  {
    if (root_fx)
      fx_destroy (root_fx);

    root_fx = preset_load 
      (gtk_object_get_data (GTK_OBJECT (menuitem), "preset_name"));

    /* We need to update configure_window if it's open. */
    if (configure_window != NULL)
      rebuild_preset_tree ();
  }
  pthread_mutex_unlock (&root_fx_mutex);
}

static gboolean popup_menu_add_dir (GtkMenu *menu, gchar *path) {
  DIR *dir;
  struct stat fileinfo;
  struct dirent *direntry;
  GList *filelist, *dirlist, *item;
  GtkWidget *menuitem;
  gchar *file;
  GString *my_path;

  filelist = dirlist = NULL;

  my_path = g_string_new (path);
  my_path = g_string_append_c (my_path, '/');

  if ((dir = opendir (path)) == NULL) {
    g_warning ("Couldn't open directory %s", path);
    return FALSE;
  } else { 
    while ((direntry = readdir (dir)) != NULL) {
      /* `file' string must be freed later */
      file = g_strconcat (my_path->str, direntry->d_name, NULL);

      /* Skip pseudo-directories */
      if (strcmp (direntry->d_name, "..") && 
	  strcmp (direntry->d_name, ".")) {
	if (stat (file, &fileinfo) == -1) {
	  g_warning ("Couldn't stat file %s", file);
	} else if (S_ISDIR(fileinfo.st_mode)) { /* A directory */
	  dirlist = g_list_append (dirlist, file);
	} else { /* A file */
	  filelist = g_list_append (filelist, file);
	}
      }
    }
  }
  
  closedir (dir);
  g_string_free (my_path, TRUE);

  if (dirlist == NULL && filelist == NULL) return FALSE;

  /* Create sub menus */
  item = g_list_first (dirlist);
  while (item) {
    GtkWidget *submenu;

    submenu = gtk_menu_new ();
    popup_menu_add_dir (GTK_MENU(submenu), (gchar *)item->data);
    gtk_widget_show (submenu);

    menuitem = gtk_menu_item_new_with_label ((gchar *)basename (item->data));
    gtk_menu_item_set_submenu (GTK_MENU_ITEM(menuitem), GTK_WIDGET(submenu));
    gtk_menu_append (GTK_MENU(menu), menuitem);
    gtk_widget_show (menuitem);

    g_free (item->data);

    item = g_list_next (item);
  }

  /* We won't need it anymore. */
  g_list_free (dirlist);

  /* Create menu items */
  item = g_list_first (filelist);
  while (item) {
    menuitem = gtk_menu_item_new_with_label 
      ((gchar *) basename ((char *) item->data));
    gtk_menu_append (GTK_MENU(menu), menuitem);

    gtk_object_set_data (GTK_OBJECT(menuitem), "preset_name", item->data);

    gtk_signal_connect (GTK_OBJECT(menuitem), "activate",
			GTK_SIGNAL_FUNC(on_popup_menu_item_activate),
			NULL);

    gtk_widget_show (menuitem);

    item = g_list_next (item);
  }

  g_list_free (filelist);

  return TRUE;
}

/* Callbacks. */

void
on_popup_menu_show                     (GtkWidget       *widget,
                                        gpointer         user_data)
{
  /* Add global presets. */
  {
    gchar *global_preset_dir = XVS_get_global_preset_dir ();
    popup_menu_add_dir (GTK_MENU(widget), global_preset_dir);
    g_free (global_preset_dir);
  }

  /* Add user presets. */
  {
    GtkWidget *submenu, *menuitem;
    gchar *user_preset_dir = XVS_get_preset_dir ();

    submenu = gtk_menu_new ();
    popup_menu_add_dir (GTK_MENU(submenu), user_preset_dir);
    gtk_widget_show (submenu);

    menuitem = gtk_menu_item_new_with_label ("User");
    gtk_menu_item_set_submenu (GTK_MENU_ITEM(menuitem), GTK_WIDGET(submenu));
    gtk_menu_append (GTK_MENU(widget), menuitem);
    gtk_widget_show (menuitem);

    g_free (user_preset_dir);
  }
}


void
on_popup_menu_configure_activate       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  XVS_configure ();
}


void
on_popup_menu_about_activate           (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  XVS_about ();
}
