/* blur.c -- a blur effect
 *
 * Copyright (C) 2001 Janusz Gregorczyk <jgregor@kki.net.pl>
 *
 * This file is part of xvs.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */
 
#include "../fx.h"
#include "../image.h"

static char* get_desc (void);
static int render (XVS_Effect *fx, XVS_Image *img, XVS_Image *img_aux,
		   gboolean beat);
static GtkWidget* configure (XVS_Effect *fx);

/* Module options. */
static const XVS_Option_Desc FXBlur_Option_Desc[] = {
  { "fade_speed", XVS_OPT_INT, { .ival = 0 } }, 
  { 0 }
};

/* Module description. */
XVS_Effect_Desc FXBlur = {
  "Filter/Blur",
  FXBlur_Option_Desc,
  0,
  
  &get_desc,
  NULL,
  &render,
  NULL,
  &configure,
};

/* Module code. */
static char *get_desc (void) {
  return "Blur";
}

static int render (XVS_Effect *fx, XVS_Image *img, XVS_Image *img_aux,
		   gboolean beat) {
  guint32 *src, *dest;
  guint32 rf, gf, bf;
  int i, chunks;
  gint fade_speed = fx->config[0].value.ival;

  rf = fade_speed << 2;
  gf = fade_speed << 10;
  bf = fade_speed << 18;

  src = img->pixels + img->w;
  dest = img_aux->pixels + img->w;

  chunks = (img->h > 2) ? img->w * (img->h - 2) : 0;

  while (chunks--) {
    guint32 rgba, r, g, b;

    rgba = src[-1];
    r = rgba & 0xff;
    g = rgba & 0xff00;
    b = rgba & 0xff0000;
    
    rgba = src[1];
    r += rgba & 0xff;
    g += rgba & 0xff00;
    b += rgba & 0xff0000;

    rgba = src[-img->w];
    r += rgba & 0xff;
    g += rgba & 0xff00;
    b += rgba & 0xff0000;
    
    rgba = src[img->w];
    r += rgba & 0xff;
    g += rgba & 0xff00;
    b += rgba & 0xff0000;
    
    r &= 0x3fc;
    g &= 0x3fc00;
    b &= 0x3fc0000;

    /* Fade out. */
    r = r > rf ? r - rf : 0;
    g = g > gf ? g - gf : 0;
    b = b > bf ? b - bf : 0;

    *dest++ = (r | g | b) >> 2;
    src++;
  }

  /* The first and the last row are a special case. */
  if (img->h >= 1) {
    dest = img_aux->pixels;
    for (i = 0; i < img->w; i++) 
      *dest++ = *(dest + img->w);
    
    dest = img_aux->pixels + img->w * (img->h - 1);
    for (i = 0; i < img->w; i++) 
      *dest++ = *(dest - img->w);
  }

  return FB_AUX;
}

static void fade_value_changed (GtkWidget *widget, gint *value) {
  *value = (int)(GTK_ADJUSTMENT(widget)->value);
}

static GtkWidget* configure (XVS_Effect *fx) {
  GtkWidget *vbox;
  gint *opt_fade_speed = &fx->config[0].value.ival;

  /* Create a new vertical box. */
  vbox = gtk_vbox_new (FALSE, 5);

  {
    GtkWidget *ahscale;
    GtkObject *aadj;

    /* Make some labels. */
    {
      GtkWidget *hbox, *label;
      hbox = gtk_hbox_new (FALSE, 5);

      label = gtk_label_new ("Slow");
      gtk_box_pack_start (GTK_BOX(hbox), label, TRUE, TRUE, 4);
      gtk_widget_show (label);

      label = gtk_label_new ("Fade speed");
      gtk_box_pack_start (GTK_BOX(hbox), label, TRUE, TRUE, 4);
      gtk_widget_show (label);

      label = gtk_label_new ("Fast");
      gtk_box_pack_start (GTK_BOX(hbox), label, TRUE, TRUE, 4);
      gtk_widget_show (label);

      gtk_box_pack_start (GTK_BOX(vbox), hbox, TRUE, TRUE, 4);
      gtk_widget_show (hbox);
    }
    
    aadj = gtk_adjustment_new (*opt_fade_speed, 0, 16, 1, 1, 0);
    ahscale = gtk_hscale_new (GTK_ADJUSTMENT(aadj));
    gtk_scale_set_draw_value (GTK_SCALE(ahscale), TRUE);
    gtk_scale_set_digits (GTK_SCALE(ahscale), 0);
    gtk_widget_show (ahscale);
    gtk_box_pack_start (GTK_BOX(vbox), ahscale, TRUE, TRUE, 0);
    
    gtk_signal_connect (GTK_OBJECT(aadj), "value-changed",
			GTK_SIGNAL_FUNC(fade_value_changed),
			opt_fade_speed);
  }

  return vbox;
}
