/* bump.c -- a bump-mapper
 *
 * Copyright (C) 2001 Janusz Gregorczyk <jgregor@kki.net.pl>
 *
 * This file is part of xvs.
 *
 * Portions of this file are derived from Blursk plugin:
 *   Blursk 1.1
 *   Copyright (c) 1999 by Steve Kirkendall <kirkenda@cs.pdx.edu>
 *   http://www.cs.pdx.edu/~kirkenda/blursk/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#include <pthread.h>

#include "../fx.h"
#include "../image.h"
#include "../widgets/color_button.h"

static char* get_desc (void);
static void init (XVS_Effect *fx);
static int render (XVS_Effect *fx, XVS_Image *img, XVS_Image *img_aux,
		   gboolean beat);
static void cleanup (XVS_Effect *fx);
static GtkWidget* configure (XVS_Effect *fx);

/* Module options. */
static const XVS_Option_Desc FXBump_Option_Desc[] = {
  { "color", XVS_OPT_INT, { .ival = 0x0010ff10 } }, 
  { 0 }
};

/* Module data. */
typedef struct {
  pthread_mutex_t mutex;
  guint32* color_map;
} FXBump_Data;

/* Module description. */
XVS_Effect_Desc FXBump = {
  "Filter/Bump",
  FXBump_Option_Desc,
  0,
  
  &get_desc,
  &init,
  &render,
  &cleanup,
  &configure,
};

/* Module code. */
static guint32 *prepare_color_map (guint32 color) {
  int i;

  guint32 *map = (guint32 *)g_malloc (256 * sizeof (guint32));
  
  for (i = 0; i < 256; i++) {
    map[i] = ((((color & 0xff) * i) & 0xff00)
            | (((color & 0xff00) * i) & 0xff0000)
            | (((color & 0xff0000) * i) & 0xff000000)) >> 8;
  }
 
  return map;
}

static char* get_desc (void) {
  return "Bump";
}

static void init (XVS_Effect *fx) {
  fx->data = g_new (FXBump_Data, 1);
  
  {
    FXBump_Data *data = (FXBump_Data *)fx->data;
    
    pthread_mutex_init (&data->mutex, NULL);
    data->color_map = NULL;
  }
}

static int render (XVS_Effect *fx, XVS_Image *img, XVS_Image *img_aux,
		   gboolean beat) {
  guint32 *dst, *src, *end;
  int delta;
  FXBump_Data *data = (FXBump_Data *)fx->data;
  gint opt_color = fx->config[0].value.ival;

  pthread_mutex_lock (&data->mutex);

  if (data->color_map == NULL) {
    data->color_map = prepare_color_map (opt_color);
  }

  /* Bump the map. */
  delta = 3 * img->w + 2;

  dst = img_aux->pixels;
  end = img_aux->pixels + delta;
  if (end > img_aux->pixels + img->h * img->w)
    end = img_aux->pixels + img->h * img->w;
  while (dst < end)
    *dst++ = data->color_map[128];

  src = img->pixels + delta;
  dst = img_aux->pixels + delta;
  end = img_aux->pixels + img->h * img->w;

  for (; dst < end; dst++, src++) {
    gint i = (256 + (src[0] & 0xff) - (src[-delta] & 0xff)) / 2;
    *dst = data->color_map[i];
  }

  pthread_mutex_unlock (&data->mutex);

  return FB_AUX;
}

static void cleanup (XVS_Effect *fx) {
  FXBump_Data *data = (FXBump_Data *)fx->data;

  pthread_mutex_destroy (&data->mutex);
  g_free (data->color_map);

  g_free (data);
}

static void color_changed (Color_Button *cb, XVS_Effect *fx) {
  FXBump_Data* data = (FXBump_Data *)fx->data;
  gint *opt_color = &fx->config[0].value.ival;

  pthread_mutex_lock (&data->mutex);

  *opt_color = cb->color;
  g_free (data->color_map);
  data->color_map = NULL;

  pthread_mutex_unlock (&data->mutex);
}    

static GtkWidget* configure (XVS_Effect *fx) {
  GtkWidget *hbox, *button, *label;
  gint *opt_color = &fx->config[0].value.ival;
  
  hbox = gtk_hbox_new (FALSE, 4);
  gtk_widget_show (hbox);
    
  label = gtk_label_new ("Bump color");
  gtk_box_pack_start (GTK_BOX(hbox), label, FALSE, FALSE, 0);
  gtk_widget_show (label);
    
  button = color_button_new ();
  color_button_set_color (COLOR_BUTTON(button), *opt_color);
  gtk_signal_connect (GTK_OBJECT(button), "color-changed", 
                      GTK_SIGNAL_FUNC(color_changed), fx);
  gtk_box_pack_start (GTK_BOX(hbox), button, FALSE, FALSE, 0);
  gtk_widget_show (button);

  return hbox;
}
