/* misc.c -- various drawing routines
 *
 * Copyright (C) 2001 Janusz Gregorczyk <jgregor@kki.net.pl>
 *
 * This file is part of xvs.
 *
 * Line drawing routine is taken from Infinity Plug-in:
 *   Infinity -- An XMMS Visualization Plug-in (version 0.2)
 *   Copyright (C) 2000 Julien Carme <julien.carme@acm.org>
 *
 * Portions of this file are derived from Jakdaw's XMMS Plugin:
 *   Jakdaw's XMMS Plugin
 *   Copyright (c) 1999, 2000 Christopher Wilson <jakdaw@usa.net>
 *   http://www.jakdaw.ucam.org/xmms/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#include <math.h>

#include "misc.h"

/* Helper functions. */

#define MIN(x, y) (((x) < (y)) ? (x) : (y))
#define MAX(x, y) (((x) > (y)) ? (x) : (y))

static __inline__ void SWAP (int *x, int *y) {
  int t = *x; *x = *y; *y = t;
}

/* Blending helpers. */

static __inline__ guint32 BLEND_ADD (guint32 a, guint32 b) {
  guint32 result, tmp;

  tmp = (a & 0xff) + (b & 0xff);
  result = MIN(tmp, 255);

  tmp = (a & 0xff00) + (b & 0xff00);
  result |= MIN(tmp >> 8, 255) << 8;

  tmp = (a & 0xff0000) + (b & 0xff0000);
  return result | (MIN(tmp >> 16, 255) << 16);
}

static __inline__ guint32 BLEND_SUB (guint32 a, guint32 b) {
  gint32 result, tmp;

  tmp = (a & 0xff) - (b & 0xff);
  result = MAX(tmp, 0);

  tmp = ((a & 0xff00) >> 8) - ((b & 0xff00) >> 8);
  result |= MAX(tmp, 0) << 8;

  tmp = ((a & 0xff0000) >> 16) - ((b & 0xff0000) >> 16);
  return result | (MAX(tmp, 0) << 16);
}

static __inline__ guint32 BLEND_AVG (guint32 a, guint32 b) {
  guint32 result, tmp;

  tmp = ((a & 0xff) + (b & 0xff)) >> 1;
  result = tmp;

  tmp = ((a & 0xff00) + (b & 0xff00)) >> 1;
  result |= tmp;

  tmp = ((a & 0xff0000) + (b & 0xff0000)) >> 1;
  return result | tmp;
}

/* Putpixels. */

static __inline__ void putpixel (XVS_Image *img, int x, int y, guint32 color) {
  if (x >= 0 && x < img->w && y >= 0 && y < img->h)
    img->pixels[y * img->w + x] = color;
}

static __inline__ void putpixel_add (XVS_Image *img, int x, int y, guint32 color) {
  if (x >= 0 && x < img->w && y >= 0 && y < img->h)
    img->pixels[y * img->w + x] = BLEND_ADD(color, img->pixels[y * img->w + x]);
}

static __inline__ void putpixel_sub (XVS_Image *img, int x, int y, guint32 color) {
  if (x >= 0 && x < img->w && y >= 0 && y < img->h)
    img->pixels[y * img->w + x] = BLEND_SUB(img->pixels[y * img->w + x], color);
}

static __inline__ void putpixel_avg (XVS_Image *img, int x, int y, guint32 color) {
  if (x >= 0 && x < img->w && y >= 0 && y < img->h)
    img->pixels[y * img->w + x] = BLEND_AVG(color, img->pixels[y * img->w + x]);
}

/* Horizontal & vertical lines. */

void draw_hline (XVS_Image *img, int y, int x1, int x2, guint32 color) {
  guint32 *dest;
  int x;

  if (x2 < x1) SWAP (&x1, &x2);

  /* Do the clipping. */
  if (x2 < 0 || x1 >= img->w || y < 0 || y >= img->h) return;

  if (x1 < 0) x1 = 0;
  if (x2 >= img->w) x2 = img->w - 1;

  dest = img->pixels + y * img->w + x1;

  for (x = x1; x <= x2; x++) {
    *dest++ = color;
  }
}

void draw_hline_blend (XVS_Image *img, int y, int x1, int x2,
		       guint32 color, int mode) {
  guint32 *dest;
  int x;

  if (x2 < x1) SWAP (&x1, &x2);

  /* Do the clipping. */
  if (x2 < 0 || x1 >= img->w || y < 0 || y >= img->h) return;

  if (x1 < 0) x1 = 0;
  if (x2 >= img->w) x2 = img->w - 1;

  dest = img->pixels + y * img->w + x1;

  switch (mode) {
  case BM_REPLACE:
    for (x = x1; x <= x2; x++) putpixel (img, x, y, color); break;
  case BM_ADDITIVE:
    for (x = x1; x <= x2; x++) putpixel_add (img, x, y, color); break;
  case BM_SUBSTRACTIVE:
    for (x = x1; x <= x2; x++) putpixel_sub (img, x, y, color); break;
  case BM_AVERAGE:
    for (x = x1; x <= x2; x++) putpixel_avg (img, x, y, color); break;
  }
}

void draw_hline_nc (XVS_Image *img, int y, int x1, int x2, guint32 color) {
  guint32 *dest;
  int x;

  if (x2 < x1) SWAP (&x1, &x2);

  dest = img->pixels + y * img->w + x1;

  for (x = x1; x <= x2; x++) {
    *dest++ = color;
  }
}

void draw_vline_nc (XVS_Image *img, int x, int y1, int y2, guint32 color) {
  guint32 *dest;
  int y;

  if (y2 < y1) SWAP (&y1, &y2);

  dest = img->pixels + y1 * img->w + x;

  for (y = y1; y <= y2; y++) {
    *dest = color;
    dest += img->w;
  }
}

/* Arbitrary lines. */

void draw_line (XVS_Image *img, int x1, int y1, int x2, int y2, guint32 color) {
  int dx, dy, cxy, dxy;
  
  /* Calculate the distances. */
  dx = abs (x1 - x2);
  dy = abs (y1 - y2);

  cxy = 0;
  if (dy > dx) {
    /* Follow Y axis */
    if (y1 > y2) {
      SWAP(&y1, &y2);
      SWAP(&x1, &x2);
    }
    
    dxy = (x1 > x2) ? -1 : 1;
    
    for (y1 = y1; y1 < y2; y1++) {
      cxy += dx;
      if (cxy >= dy) {
	x1 += dxy;
	cxy -= dy;
      }
      putpixel (img, x1, y1, color);
    }		
  } else {
    /* Follow X axis */
    if (x1 > x2) {
      SWAP(&x1, &x2);
      SWAP(&y1, &y2);
    }
    
    dxy = (y1 > y2) ? -1 : 1;
    
    for (x1 = x1; x1 < x2; x1++) {
      cxy += dy;
      if (cxy >= dx) {
	y1 += dxy;
	cxy -= dx;
      }
      putpixel (img, x1, y1, color);
    }		
  }
}

/* Circles. */

void draw_circle (XVS_Image *img, int x0, int y0, int radius,
		  guint32 color, int mode) {
  /* Perform initialization. */
  int d = 3 - (2 * radius);
  int x = 0;
  int y = radius;

  do {
    draw_hline_blend (img, y0 + x, x0 - y, x0 + y, color, mode);
    if (x) draw_hline_blend (img, y0 - x, x0 - y, x0 + y, color, mode);

    if (d < 0) {
      d += (4 * x) + 6;
    } else {
      if (x != y) {
	draw_hline_blend (img, y0 + y, x0 - x, x0 + x, color, mode);
	if (y) draw_hline_blend (img, y0 - y, x0 - x, x0 + x, color, mode);
      }

      d += 4 * (x - y) + 10;
      y--;
    }
  } while (x++ < y);
}

/* Blits and filters. */

void table_blit (guint32 *dest, guint32 *src, guint32 *tab, guint tabsize) {
  guint32 *tptr = tab;

  while (tabsize--)
    *dest++ = src[*tptr++];
}

void table_blit_bf (guint32 *dest, guint32 *src, guint32 *tab, guint tabsize) {
  guint32 *tptr = tab;

  while (tabsize--) {
    guint32 rgba, r, g, b;

    rgba = src[*tptr++];
    r = rgba & 0xff;
    g = rgba & 0xff00;
    b = rgba & 0xff0000;
    
    rgba = src[*tptr++];
    r += rgba & 0xff;
    g += rgba & 0xff00;
    b += rgba & 0xff0000;

    rgba = src[*tptr++];
    r += rgba & 0xff;
    g += rgba & 0xff00;
    b += rgba & 0xff0000;
    
    rgba = src[*tptr++];
    r += rgba & 0xff;
    g += rgba & 0xff00;
    b += rgba & 0xff0000;
    
    r &= 0x3fc;
    g &= 0x3fc00;
    b &= 0x3fc0000;
    
    *dest++ = (r | g | b) >> 2;
  }
}
