/* particle.c -- a moving particle
 *
 * Copyright (C) 2001 Janusz Gregorczyk <jgregor@kki.net.pl>
 *
 * This file is part of xvs.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#include <math.h>
#include <stdlib.h>

#include "misc.h"
#include "../fx.h"
#include "../image.h"
#include "../widgets/color_button.h"

static char* get_desc (void);
static void init (XVS_Effect *fx);
static int render (XVS_Effect *fx, XVS_Image *img, XVS_Image *img_aux,
		   gboolean beat);
static void cleanup (XVS_Effect *fx);
static GtkWidget* configure (XVS_Effect *fx);

/* Module options. */
static const XVS_Option_Desc FXParticle_Option_Desc[] = {
  { "color", XVS_OPT_INT, { .ival = 0x0000ffff } },
  { "blending_mode", XVS_OPT_INT, { .ival = BM_ADDITIVE } },
  { "size", XVS_OPT_INT, { .ival = 5 } },
  { "radius", XVS_OPT_FLOAT, { .fval = 0.3 } },
  { "acceleration", XVS_OPT_FLOAT, { .fval = -0.05 } },
  { "muffle", XVS_OPT_FLOAT, { .fval = 0.95 } },
  { 0 }
};  

#define GET_COLOR(fx)		(&fx->config[0].value.ival)
#define GET_BLENDING_MODE(fx)	(&fx->config[1].value.ival)
#define GET_SIZE(fx) 		(&fx->config[2].value.ival)
#define GET_RADIUS(fx) 		(&fx->config[3].value.fval)
#define GET_ACCELERATION(fx)	(&fx->config[4].value.fval)
#define GET_MUFFLE(fx) 		(&fx->config[5].value.fval)

/* Module data. */
typedef struct {
  gfloat x, y;			/* Current position */
  gfloat x0, y0;		/* `target' position */
  gfloat vx, vy;		/* Velocity  */
  gfloat ax, ay;		/* Accelleration */
} FXParticle_Data;

/* Module description. */
XVS_Effect_Desc FXParticle = {
  "Render/Moving Particle",
  FXParticle_Option_Desc,
  0,
  
  &get_desc,
  &init,
  &render,
  &cleanup,
  &configure
};

/* Module code. */
static char* get_desc (void) {
  return "Moving particle";
}

static void init (XVS_Effect *fx) {
  fx->data = g_new (FXParticle_Data, 1);
  
  {
    FXParticle_Data *data = (FXParticle_Data *)fx->data;
    
    data->x = data->y = 0.0;
    data->x0 = data->y0 = 0.0;
    data->vx = data->vy = 0.0;
    data->ax = data->ay = 0.0;
  }
}

static int render (XVS_Effect *fx, XVS_Image *img, XVS_Image *img_aux,
		   gboolean beat) {
  FXParticle_Data *data = (FXParticle_Data *)fx->data;
  int img_size = (img->w < img->h) ? img->w : img->h;

  if (beat) {
    double angle = 2.0 * M_PI * rand () / RAND_MAX;

    data->x0 = (img_size / 2) * (*GET_RADIUS(fx)) * sin (angle);
    data->y0 = (img_size / 2) * (*GET_RADIUS(fx)) * cos (angle);
  }

  data->ax = (*GET_ACCELERATION(fx)) * (data->x - data->x0);
  data->ay = (*GET_ACCELERATION(fx)) * (data->y - data->y0);

  data->vx = (*GET_MUFFLE(fx)) * data->vx + data->ax;
  data->vy = (*GET_MUFFLE(fx)) * data->vy + data->ay;

  data->x += data->vx;
  data->y += data->vy;

  draw_circle (img, 
	       (img->w >> 1) + (int)data->x,
	       (img->h >> 1) + (int)data->y, 
	       *GET_SIZE(fx) * img_size / 200,
	       *GET_COLOR(fx), *GET_BLENDING_MODE(fx));

  return FB_SINGLE;
}

static void cleanup (XVS_Effect *fx) {
  FXParticle_Data *data = (FXParticle_Data *)fx->data;
  g_free (data);
}

static void blending_mode_sel (GtkObject *item, int *value) {
  *value = (int)gtk_object_get_user_data (item);
}

static void size_value_changed (GtkWidget *widget, int *value) {
  *value = (int)GTK_ADJUSTMENT(widget)->value;
}

static void radius_value_changed (GtkWidget *widget, gfloat *value) {
  *value = GTK_ADJUSTMENT(widget)->value;
}

static void color_changed (Color_Button *cb, gint *value) {
  *value = cb->color;
}    

static GtkWidget* configure (XVS_Effect *fx) {
  GtkWidget *vbox, *label, *ahscale;
  GtkObject *aadj;

  vbox = gtk_vbox_new (FALSE, 5);

  /* Blending mode. */
  {
    const char *bm_desc[] = { 
      "Replace", "Additive", "Substractive", "50%", NULL };
    const int bm_value[] = { 
      BM_REPLACE, BM_ADDITIVE, BM_SUBSTRACTIVE, BM_AVERAGE };

    GtkWidget *menu, *optionmenu, *hbox;
    int i;

    hbox = gtk_hbox_new (FALSE, 4);
    gtk_box_pack_start (GTK_BOX(vbox), hbox, FALSE, FALSE, 4);
    gtk_widget_show (hbox);
    
    label = gtk_label_new ("Blending mode");
    gtk_box_pack_start (GTK_BOX(hbox), label, FALSE, FALSE, 0);
    gtk_widget_show (label);

    menu = gtk_menu_new ();

    for (i = 0; bm_desc[i]; i++) {
      GtkWidget *item;

      item = gtk_menu_item_new_with_label (bm_desc[i]);
      gtk_widget_show(item);

      gtk_object_set_user_data (GTK_OBJECT(item), (gpointer)bm_value[i]);
      gtk_signal_connect (GTK_OBJECT(item),
			  "activate", GTK_SIGNAL_FUNC(blending_mode_sel),
			  GET_BLENDING_MODE(fx));
      gtk_menu_append (GTK_MENU(menu), item);
    }
    gtk_menu_set_active(GTK_MENU(menu), *GET_BLENDING_MODE(fx));
    gtk_widget_show (menu);
	
    optionmenu = gtk_option_menu_new ();
    gtk_option_menu_set_menu (GTK_OPTION_MENU(optionmenu), menu);
    gtk_widget_show (optionmenu);
    gtk_box_pack_start (GTK_BOX(hbox), optionmenu, FALSE, FALSE, 0);
  }

  /* Size */  
  {
    label = gtk_label_new ("Particle size");
    gtk_box_pack_start (GTK_BOX(vbox), label, FALSE, FALSE, 4);
    gtk_widget_show (label);
    
    aadj = gtk_adjustment_new (*GET_SIZE(fx), 0, 100, 1, 10, 0);
    ahscale = gtk_hscale_new (GTK_ADJUSTMENT(aadj));
    gtk_scale_set_draw_value (GTK_SCALE(ahscale), TRUE);
    gtk_scale_set_value_pos (GTK_SCALE(ahscale), GTK_POS_LEFT);
    gtk_scale_set_digits (GTK_SCALE(ahscale), 0);
    gtk_widget_show (ahscale);
    gtk_box_pack_start (GTK_BOX(vbox), ahscale, TRUE, TRUE, 4);
    
    gtk_signal_connect (GTK_OBJECT(aadj), "value-changed",
			GTK_SIGNAL_FUNC(size_value_changed),
			GET_SIZE(fx));
  }

  /* Position */  
  {
    label = gtk_label_new ("Distance from the center (% of window)");
    gtk_box_pack_start (GTK_BOX(vbox), label, FALSE, FALSE, 4);
    gtk_widget_show (label);
    
    aadj = gtk_adjustment_new (*GET_RADIUS(fx), 0, 1, 0.01, 0.1, 0);
    ahscale = gtk_hscale_new (GTK_ADJUSTMENT(aadj));
    gtk_scale_set_draw_value (GTK_SCALE(ahscale), TRUE);
    gtk_scale_set_value_pos (GTK_SCALE(ahscale), GTK_POS_LEFT);
    gtk_scale_set_digits (GTK_SCALE(ahscale), 2);
    gtk_widget_show (ahscale);
    gtk_box_pack_start (GTK_BOX(vbox), ahscale, TRUE, TRUE, 4);
    
    gtk_signal_connect (GTK_OBJECT(aadj), "value-changed",
			GTK_SIGNAL_FUNC(radius_value_changed),
			GET_RADIUS(fx));
  }

  /* Colour selection */
  { 
    GtkWidget *hbox, *button;

    hbox = gtk_hbox_new (FALSE, 4);
    gtk_box_pack_start (GTK_BOX(vbox), hbox, FALSE, FALSE, 4);
    gtk_widget_show (hbox);
    
    label = gtk_label_new ("Colour");
    gtk_box_pack_start (GTK_BOX(hbox), label, FALSE, FALSE, 0);
    gtk_widget_show (label);

    button = color_button_new ();
    color_button_set_color (COLOR_BUTTON(button), *GET_COLOR(fx));
    gtk_signal_connect (GTK_OBJECT(button), "color-changed", 
			GTK_SIGNAL_FUNC(color_changed), GET_COLOR(fx));
    gtk_box_pack_start (GTK_BOX(hbox), button, FALSE, FALSE, 0);
    gtk_widget_show (button);
  }

  return vbox;
}
