/* ring.c -- a ring scope plotter
 *
 * Copyright (C) 2001 Janusz Gregorczyk <jgregor@kki.net.pl>
 *
 * This file is part of xvs.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#include <stdlib.h>
#include <math.h>

#include "misc.h"
#include "../fx.h"
#include "../image.h"
#include "../widgets/color_button.h"

static char *get_desc (void);
static int render (XVS_Effect *fx, XVS_Image *img, XVS_Image *img_aux,
		   gboolean beat);
static GtkWidget* configure (XVS_Effect *fx);

enum { RING_COLOR_SOLID = 0, RING_COLOR_RANDOM = 1 };
enum { POSITION_TOP = 0, POSITION_LEFT = 1, 
       POSITION_MIDDLE = 2, 
       POSITION_BOTTOM = 3, POSITION_RIGHT = 4 };

/* Module options. */
static const XVS_Option_Desc FXRing_Option_Desc[] = {
  { "color_type", XVS_OPT_INT, { .ival = RING_COLOR_SOLID } },
  { "color", XVS_OPT_INT, { .ival = 0x00ffffff } },
  { "amplitude", XVS_OPT_INT, { .ival = 33 } },
  { "position", XVS_OPT_INT, { .ival = POSITION_MIDDLE } },
  { 0 }
};  

#define GET_COLOR_TYPE(fx)	(&fx->config[0].value.ival)
#define GET_COLOR(fx) 		(&fx->config[1].value.ival)
#define GET_AMPLITUDE(fx) 	(&fx->config[2].value.ival)
#define GET_POSITION(fx) 	(&fx->config[3].value.ival)

/* Module description. */
XVS_Effect_Desc FXRing = {
  "Render/Ring",
  FXRing_Option_Desc,
  0,
  
  &get_desc,
  NULL,
  &render,
  NULL,
  &configure
};

/* Module code. */
static char *get_desc (void) {
  return "Ring";
}

static int render (XVS_Effect *fx, XVS_Image *img, XVS_Image *img_aux,
		   gboolean beat) {
  guint32 color;
  int x, y, oldx, oldy, centerx, centery;
  int i, size, r;

  /* Decide on which color to use. */
  switch (*GET_COLOR_TYPE(fx)) {
  case RING_COLOR_SOLID: color = *GET_COLOR(fx); break;
  case RING_COLOR_RANDOM: color = random (); break;
  }

  /* Decide on the position and direction. */
  switch (*GET_POSITION(fx)) {
  case POSITION_TOP: centerx = img->w / 2; centery = img->h / 4; break;
  case POSITION_MIDDLE: centerx = img->w / 2; centery = img->h / 2; break;
  case POSITION_BOTTOM: centerx = img->w / 2; centery = 3 * img->h / 4; break;
  case POSITION_LEFT: centerx = img->w / 4; centery = img->h / 2; break;
  case POSITION_RIGHT: centerx = 3 * img->w / 4; centery = img->h / 2; break;
  }

  if (img->w > img->h)
    size = img->w;
  else
    size = img->h;

  /* Draw. */
  for (i = 0; i < 513; i++) {
    r = (*GET_AMPLITUDE(fx) * size / 100);
    r += (r * XVS_pcm_data[0][i & 511] / 3) >> 15;

    x = centerx + (int)(r * sin (2 * 3.14 * i / 512.0)); 
    y = centery + (int)(r * cos (2 * 3.14 * i / 512.0)); 

    if (i != 0)
      draw_line (img, oldx, oldy, x, y, color);

    oldx = x;
    oldy = y;
  }

  return FB_SINGLE;
}

static void amp_value_changed (GtkWidget *w, int *val) {
  *val = (int)GTK_ADJUSTMENT(w)->value;
}

static void color_type_radio_toggled (GtkToggleButton *b, XVS_Effect *fx) {
  if (gtk_toggle_button_get_active (b))
    *GET_COLOR_TYPE(fx) = (int)gtk_object_get_user_data (GTK_OBJECT(b));
}

static void on_button_clicked (GtkWidget *widget, XVS_Effect *fx) {
  *GET_POSITION(fx) = (int)gtk_object_get_user_data (GTK_OBJECT(widget));
}

static void color_changed (Color_Button *cb, gint *value) {
  *value = cb->color;
}    

static GtkWidget* configure (XVS_Effect *fx) {
  GtkWidget *vbox, *hbox, *button, *label, *radio, *frame, *vvbox;

  vbox = gtk_vbox_new (FALSE, 5);

  frame = gtk_frame_new ("Size and position");
  gtk_box_pack_start (GTK_BOX(vbox), frame, FALSE, FALSE, 0);
  gtk_widget_show (frame);
  
  vvbox = gtk_vbox_new (FALSE, 5);
  gtk_container_border_width (GTK_CONTAINER(vvbox), 4);
  gtk_container_add (GTK_CONTAINER(frame), vvbox);
  gtk_widget_show (vvbox);

  /* Ring position */
  {
    GtkWidget *table, *arrow;

    table = gtk_table_new (3, 3, FALSE);

    /* Up */
    button = gtk_button_new ();
    arrow = gtk_arrow_new (GTK_ARROW_UP, GTK_SHADOW_NONE);
    gtk_container_add (GTK_CONTAINER(button), arrow);
    gtk_widget_show (arrow);
    gtk_object_set_user_data (GTK_OBJECT(button), (gpointer)POSITION_TOP);
    gtk_signal_connect (GTK_OBJECT(button), "clicked",
			GTK_SIGNAL_FUNC(on_button_clicked), (gpointer)fx);
    gtk_table_attach_defaults (GTK_TABLE(table), button, 1, 2, 0, 1);
    gtk_widget_show (button);

    /* Down */
    button = gtk_button_new ();
    arrow = gtk_arrow_new (GTK_ARROW_DOWN, GTK_SHADOW_NONE);
    gtk_container_add (GTK_CONTAINER(button), arrow);
    gtk_widget_show (arrow);
    gtk_object_set_user_data (GTK_OBJECT(button), (gpointer)POSITION_BOTTOM);
    gtk_signal_connect (GTK_OBJECT(button), "clicked",
			GTK_SIGNAL_FUNC(on_button_clicked), (gpointer)fx);
    gtk_table_attach_defaults (GTK_TABLE(table), button, 1, 2, 2, 3);
    gtk_widget_show (button);

    /* Left */
    button = gtk_button_new ();
    arrow = gtk_arrow_new (GTK_ARROW_LEFT, GTK_SHADOW_NONE);
    gtk_container_add (GTK_CONTAINER(button), arrow);
    gtk_widget_show (arrow);
    gtk_object_set_user_data (GTK_OBJECT(button), (gpointer)POSITION_LEFT);
    gtk_signal_connect (GTK_OBJECT(button), "clicked",
			GTK_SIGNAL_FUNC(on_button_clicked), (gpointer)fx);
    gtk_table_attach_defaults (GTK_TABLE(table), button, 0, 1, 1, 2);
    gtk_widget_show (button);

    /* Right */
    button = gtk_button_new ();
    arrow = gtk_arrow_new (GTK_ARROW_RIGHT, GTK_SHADOW_NONE);
    gtk_container_add (GTK_CONTAINER(button), arrow);
    gtk_widget_show (arrow);
    gtk_object_set_user_data (GTK_OBJECT(button), (gpointer)POSITION_RIGHT);
    gtk_signal_connect (GTK_OBJECT(button), "clicked",
			GTK_SIGNAL_FUNC(on_button_clicked), (gpointer)fx);
    gtk_table_attach_defaults (GTK_TABLE(table), button, 2, 3, 1, 2);
    gtk_widget_show (button);

    /* Middle */
    button = gtk_button_new_with_label ("Center");
    gtk_object_set_user_data (GTK_OBJECT(button), (gpointer)POSITION_MIDDLE);
    gtk_signal_connect (GTK_OBJECT(button), "clicked",
			GTK_SIGNAL_FUNC(on_button_clicked), (gpointer)fx);
    gtk_table_attach_defaults (GTK_TABLE(table), button, 1, 2, 1, 2);
    gtk_widget_show (button);

    gtk_box_pack_start (GTK_BOX(vvbox), table, FALSE, FALSE, 0);
    gtk_widget_show (table);
  }

  /* Ring amplitude */  
  {
    GtkWidget *ahscale;
    GtkObject *aadj;

    label = gtk_label_new ("Ring amplitude (% of window)");
    gtk_widget_show (label);
    gtk_box_pack_start (GTK_BOX(vvbox), label, FALSE, FALSE, 4);
    
    aadj = gtk_adjustment_new (*GET_AMPLITUDE(fx), 10, 100, 1, 10, 0);
    ahscale = gtk_hscale_new (GTK_ADJUSTMENT(aadj));
    gtk_scale_set_draw_value (GTK_SCALE(ahscale), TRUE);
    gtk_scale_set_value_pos (GTK_SCALE(ahscale), GTK_POS_LEFT);
    gtk_scale_set_digits (GTK_SCALE(ahscale), 0);
    gtk_widget_show (ahscale);
    gtk_box_pack_start (GTK_BOX(vvbox), ahscale, TRUE, TRUE, 4);
    
    gtk_signal_connect (GTK_OBJECT(aadj), "value-changed",
			GTK_SIGNAL_FUNC(amp_value_changed),
			GET_AMPLITUDE(fx));
  }
  
  /* Colour selection */
  {
    frame = gtk_frame_new ("Colour");
    gtk_box_pack_start (GTK_BOX(vbox), frame, FALSE, FALSE, 0);
    gtk_widget_show (frame);

    vvbox = gtk_vbox_new (FALSE, 5);
    gtk_container_border_width (GTK_CONTAINER(vvbox), 4);
    gtk_container_add (GTK_CONTAINER(frame), vvbox);
    gtk_widget_show (vvbox);
    
    radio = gtk_radio_button_new (NULL);
    gtk_box_pack_start (GTK_BOX(vvbox), radio, FALSE, FALSE, 0);
    gtk_widget_show (radio);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(radio), 
				  GET_COLOR_TYPE(fx) == RING_COLOR_SOLID);
    gtk_object_set_user_data (GTK_OBJECT(radio), (gpointer)RING_COLOR_SOLID);
    gtk_signal_connect (GTK_OBJECT(radio), "toggled", 
			GTK_SIGNAL_FUNC(color_type_radio_toggled), (gpointer)fx);
    
    hbox = gtk_hbox_new (FALSE, 4);
    gtk_container_add (GTK_CONTAINER(radio), hbox);
    gtk_widget_show (hbox);
    
    label = gtk_label_new ("Solid Colour");
    gtk_box_pack_start (GTK_BOX(hbox), label, FALSE, FALSE, 4);
    gtk_widget_show (label);

    button = color_button_new ();
    color_button_set_color (COLOR_BUTTON(button), *GET_COLOR(fx));
    gtk_signal_connect (GTK_OBJECT(button), "color-changed", 
			GTK_SIGNAL_FUNC(color_changed), GET_COLOR(fx));
    gtk_box_pack_start (GTK_BOX(hbox), button, FALSE, FALSE, 0);
    gtk_widget_show (button);
    
    radio = gtk_radio_button_new_with_label 
      (gtk_radio_button_group(GTK_RADIO_BUTTON(radio)), "Random Colors");
    gtk_box_pack_start (GTK_BOX(vvbox), radio, FALSE, FALSE, 0);
    gtk_widget_show (radio);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(radio), 
				  *GET_COLOR_TYPE(fx) == RING_COLOR_RANDOM);
    gtk_object_set_user_data (GTK_OBJECT(radio), (gpointer)RING_COLOR_RANDOM);
    gtk_signal_connect (GTK_OBJECT(radio), "toggled", 
			GTK_SIGNAL_FUNC(color_type_radio_toggled), (gpointer)fx);
  }  

  return vbox;
}
