/* scope.c -- a simple scope plotter
 *
 * Copyright (C) 2001 Janusz Gregorczyk <jgregor@kki.net.pl>
 * 
 * This file is part of xvs.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#include <stdlib.h>

#include "misc.h"
#include "../fx.h"
#include "../image.h"
#include "../widgets/color_button.h"

static char *get_desc (void);
static int render (XVS_Effect *fx, XVS_Image *img, XVS_Image *img_aux,
		   gboolean beat);
static GtkWidget* configure (XVS_Effect *fx);

enum { SCOPE_VERTICAL = 0, SCOPE_HORIZONTAL = 1 };
enum { SCOPE_COLOR_SOLID = 0, SCOPE_COLOR_RANDOM = 1 };
enum { POINTS = 0, LINES = 1, SOLID = 2 };
enum { POSITION_TOP = 0, POSITION_LEFT = 0, 
       POSITION_MIDDLE = 1, 
       POSITION_BOTTOM = 2, POSITION_RIGHT = 2 };

/* Module options. */
static const XVS_Option_Desc FXScope_Option_Desc[] = {
  { "color_type", XVS_OPT_INT, { .ival = SCOPE_COLOR_SOLID } }, 
  { "color", XVS_OPT_INT, { .ival = 0x00ffffff } }, 
  { "orientation", XVS_OPT_INT, { .ival = SCOPE_HORIZONTAL } }, 
  { "mode", XVS_OPT_INT, { .ival = LINES } }, 
  { "amplitude", XVS_OPT_INT, { .ival = 33 } }, 
  { "position", XVS_OPT_INT, { .ival = POSITION_MIDDLE } }, 
  { 0 }
};

#define GET_COLOR_TYPE(fx)	(&fx->config[0].value.ival)
#define GET_COLOR(fx) 		(&fx->config[1].value.ival)
#define GET_ORIENTATION(fx) 	(&fx->config[2].value.ival)
#define GET_MODE(fx) 		(&fx->config[3].value.ival)
#define GET_AMPLITUDE(fx) 	(&fx->config[4].value.ival)
#define GET_POSITION(fx) 	(&fx->config[5].value.ival)

/* Module description. */
XVS_Effect_Desc FXScope = {
  "Render/Scope",
  FXScope_Option_Desc,
  0,
  
  &get_desc,
  NULL,
  &render,
  NULL,
  &configure
};

/* Module code. */
static char *get_desc (void) {
  return "Scope";
}

static int render (XVS_Effect *fx, XVS_Image *img, XVS_Image *img_aux,
		   gboolean beat) {
  guint32 color;
  int x, y, center, i, dist, prev;

  /* Decide on which color to use. */
  switch (*GET_COLOR_TYPE(fx)) {
  case SCOPE_COLOR_SOLID: color = *GET_COLOR(fx); break;
  case SCOPE_COLOR_RANDOM: color = random (); break;
  }

  /* Decide on the position and direction */
  switch (*GET_ORIENTATION(fx)) { 
  case SCOPE_HORIZONTAL: center = img->h; break;
  case SCOPE_VERTICAL: center = img->w; break;
  }

  switch (*GET_POSITION(fx)) {
  case POSITION_TOP: center /= 4; break;
  case POSITION_MIDDLE: center /= 2; break;
  case POSITION_BOTTOM: center = 3 * center / 4; break;
  }

  /* Draw */
  prev = center;
  dist = (*GET_ORIENTATION(fx) == SCOPE_HORIZONTAL) ? img->w : img->h;

  for (i = 0; i < dist; i++) {
    if (*GET_ORIENTATION(fx) == SCOPE_HORIZONTAL) {
      x = i;

      y = (*GET_AMPLITUDE(fx) * XVS_pcm_data[0][i & 511] * img->h / 100) >> 15;
      y += center ;

      if (y < 0) y = 0;
      if (y >= img->h) y = img->h - 1;
    } else {
      y = i;

      x = (*GET_AMPLITUDE(fx) * XVS_pcm_data[0][i & 511] * img->w / 100) >> 15;
      x += center;

      if (x < 0) x = 0;
      if (x >= img->w) x = img->w - 1;    
    }

    switch (*GET_MODE(fx)) {
    case POINTS: 
      *(img->pixels + y * img->w + x) = color;
      break;
    case LINES:
      if (*GET_ORIENTATION(fx) == SCOPE_HORIZONTAL)
        draw_vline_nc (img, x, y, prev, color);
      else
        draw_hline_nc (img, y, x, prev, color);
      break;    
    case SOLID:
      if (*GET_ORIENTATION(fx) == SCOPE_HORIZONTAL)
        draw_vline_nc (img, x, y, center, color);
      else
        draw_hline_nc (img, y, x, center, color);
      break;
    }

    if (*GET_ORIENTATION(fx) == SCOPE_HORIZONTAL) {
      prev = y;
    } else {
      prev = x;
    }
  }

  return FB_SINGLE;
}

static void plotter_radio_toggled (GtkToggleButton *b, XVS_Effect *fx) {
  if (gtk_toggle_button_get_active (b))
    *GET_MODE(fx) = (int)gtk_object_get_user_data (GTK_OBJECT(b));
}

static void orientation_radio_toggled (GtkToggleButton *b, XVS_Effect *fx) {
  if (gtk_toggle_button_get_active (b))
    *GET_ORIENTATION(fx) = (int)gtk_object_get_user_data (GTK_OBJECT(b));
}

static void position_radio_toggled (GtkToggleButton *b, XVS_Effect *fx) {
  if (gtk_toggle_button_get_active (b))
    *GET_POSITION(fx) = (int)gtk_object_get_user_data (GTK_OBJECT(b));
}

static void color_type_radio_toggled (GtkToggleButton *b, XVS_Effect *fx) {
  if (gtk_toggle_button_get_active (b))
    *GET_COLOR_TYPE(fx) = (int)gtk_object_get_user_data (GTK_OBJECT(b));
}

static void amp_value_changed (GtkWidget *w, int *val) {
  *val = (int)GTK_ADJUSTMENT(w)->value;
}

static void color_changed (Color_Button *cb, gint *value) {
  *value = cb->color;
}    

static GtkWidget* configure (XVS_Effect *fx) {
  GtkWidget *vbox, *hbox, *button, *label, *radio, *frame, *vvbox;

  vbox = gtk_vbox_new (FALSE, 5);

  /* Plotter mode */  
  frame = gtk_frame_new ("Plotter mode");
  gtk_box_pack_start (GTK_BOX(vbox), frame, FALSE, FALSE, 0);
  gtk_widget_show (frame);
  
  vvbox = gtk_vbox_new (FALSE, 5);
  gtk_container_add (GTK_CONTAINER(frame), vvbox);
  gtk_widget_show (vvbox);

  {
    hbox = gtk_hbox_new (FALSE, 4);
    gtk_widget_show (hbox);
    gtk_container_add (GTK_CONTAINER(vvbox), hbox);

    radio = gtk_radio_button_new_with_label (NULL, "Points");
    gtk_box_pack_start (GTK_BOX(hbox), radio, FALSE, FALSE, 4);
    gtk_widget_show (radio);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(radio), 
				  *GET_MODE(fx) == POINTS);
    gtk_object_set_user_data (GTK_OBJECT(radio), (gpointer)POINTS);
    gtk_signal_connect (GTK_OBJECT(radio), "toggled", 
			GTK_SIGNAL_FUNC(plotter_radio_toggled), (gpointer)fx);
  
    radio = gtk_radio_button_new_with_label 
      (gtk_radio_button_group(GTK_RADIO_BUTTON(radio)), "Lines");
    gtk_box_pack_start (GTK_BOX(hbox), radio, FALSE, FALSE, 4);
    gtk_widget_show (radio);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(radio), 
				  *GET_MODE(fx) == LINES);
    gtk_object_set_user_data (GTK_OBJECT(radio), (gpointer)LINES);

    gtk_signal_connect (GTK_OBJECT(radio), "toggled", 
			GTK_SIGNAL_FUNC(plotter_radio_toggled), (gpointer)fx);

    radio = gtk_radio_button_new_with_label 
      (gtk_radio_button_group(GTK_RADIO_BUTTON(radio)), "Solid");
    gtk_box_pack_start (GTK_BOX(hbox), radio, FALSE, FALSE, 4);
    gtk_widget_show (radio);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(radio), 
				  *GET_MODE(fx) == SOLID);
    gtk_object_set_user_data (GTK_OBJECT(radio), (gpointer)SOLID);
    gtk_signal_connect (GTK_OBJECT(radio), "toggled", 
			GTK_SIGNAL_FUNC(plotter_radio_toggled), (gpointer)fx);
  }

  frame = gtk_frame_new ("Size and position");
  gtk_box_pack_start (GTK_BOX(vbox), frame, FALSE, FALSE, 0);
  gtk_widget_show (frame);
  
  vvbox = gtk_vbox_new (FALSE, 5);
  gtk_container_add (GTK_CONTAINER(frame), vvbox);
  gtk_widget_show (vvbox);

  /* Scope orientation */  
  {
    hbox = gtk_hbox_new (FALSE, 4);
    gtk_container_add (GTK_CONTAINER(vvbox), hbox);
    gtk_widget_show (hbox);

    radio = gtk_radio_button_new_with_label (NULL, "Horizontal");
    gtk_box_pack_start (GTK_BOX(hbox), radio, FALSE, FALSE, 4);
    gtk_widget_show (radio);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(radio), 
				  *GET_ORIENTATION(fx) == SCOPE_HORIZONTAL);
    gtk_object_set_user_data (GTK_OBJECT(radio), (gpointer)SCOPE_HORIZONTAL);
    gtk_signal_connect (GTK_OBJECT(radio), "toggled", 
			GTK_SIGNAL_FUNC(orientation_radio_toggled), (gpointer)fx);
  
    radio = gtk_radio_button_new_with_label 
      (gtk_radio_button_group(GTK_RADIO_BUTTON(radio)), "Vertical");
    gtk_box_pack_start (GTK_BOX(hbox), radio, FALSE, FALSE, 4);
    gtk_widget_show (radio);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(radio), 
				  *GET_ORIENTATION(fx) == SCOPE_VERTICAL);
    gtk_object_set_user_data (GTK_OBJECT(radio), (gpointer)SCOPE_VERTICAL);
    gtk_signal_connect (GTK_OBJECT(radio), "toggled", 
			GTK_SIGNAL_FUNC(orientation_radio_toggled), (gpointer)fx);
  }

  /* Scope position */
  {
    hbox = gtk_hbox_new (FALSE, 4);
    gtk_container_add (GTK_CONTAINER(vvbox), hbox);
    gtk_widget_show (hbox);

    radio = gtk_radio_button_new_with_label (NULL, "Top/Left");
    gtk_box_pack_start (GTK_BOX(hbox), radio, FALSE, FALSE, 4);
    gtk_widget_show (radio);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(radio), 
				  *GET_POSITION(fx) == POSITION_TOP);
    gtk_object_set_user_data (GTK_OBJECT(radio), (gpointer)POSITION_TOP);
    gtk_signal_connect (GTK_OBJECT(radio), "toggled", 
			GTK_SIGNAL_FUNC(position_radio_toggled), (gpointer)fx);
  
    radio = gtk_radio_button_new_with_label 
      (gtk_radio_button_group(GTK_RADIO_BUTTON(radio)), "Middle");
    gtk_box_pack_start (GTK_BOX(hbox), radio, FALSE, FALSE, 4);
    gtk_widget_show (radio);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(radio), 
				  *GET_POSITION(fx) == POSITION_MIDDLE);
    gtk_object_set_user_data (GTK_OBJECT(radio), (gpointer)POSITION_MIDDLE);
    gtk_signal_connect (GTK_OBJECT(radio), "toggled", 
			GTK_SIGNAL_FUNC(position_radio_toggled), (gpointer)fx);

    radio = gtk_radio_button_new_with_label 
      (gtk_radio_button_group(GTK_RADIO_BUTTON(radio)), "Right/Bottom");
    gtk_box_pack_start (GTK_BOX(hbox), radio, FALSE, FALSE, 4);
    gtk_widget_show (radio);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(radio), 
				  *GET_POSITION(fx) == POSITION_BOTTOM);
    gtk_object_set_user_data (GTK_OBJECT(radio), (gpointer)POSITION_BOTTOM);
    gtk_signal_connect (GTK_OBJECT(radio), "toggled", 
			GTK_SIGNAL_FUNC(position_radio_toggled), (gpointer)fx);
  }

  /* Scope amplitude */  
  {
    GtkWidget *ahscale;
    GtkObject *aadj;

    label = gtk_label_new ("Scope amplitude (% of window)");
    gtk_box_pack_start (GTK_BOX(vvbox), label, FALSE, FALSE, 4);
    gtk_widget_show (label);
    
    aadj = gtk_adjustment_new (*GET_AMPLITUDE(fx), 10, 100, 1, 10, 0);
    ahscale = gtk_hscale_new (GTK_ADJUSTMENT(aadj));
    gtk_scale_set_draw_value (GTK_SCALE(ahscale), TRUE);
    gtk_scale_set_value_pos (GTK_SCALE(ahscale), GTK_POS_LEFT);
    gtk_scale_set_digits (GTK_SCALE(ahscale), 0);
    gtk_widget_show (ahscale);
    gtk_box_pack_start (GTK_BOX(vvbox), ahscale, TRUE, TRUE, 4);
    
    gtk_signal_connect (GTK_OBJECT(aadj), "value-changed",
			GTK_SIGNAL_FUNC(amp_value_changed),
			GET_AMPLITUDE(fx));
  }
  
  /* Colour selection */
  {
    frame = gtk_frame_new ("Colour");
    gtk_box_pack_start (GTK_BOX(vbox), frame, FALSE, FALSE, 0);
    gtk_widget_show (frame);

    vvbox = gtk_vbox_new (FALSE, 5);
    gtk_container_border_width (GTK_CONTAINER(vvbox), 4);
    gtk_container_add (GTK_CONTAINER(frame), vvbox);
    gtk_widget_show (vvbox);
    
    radio = gtk_radio_button_new (NULL);
    gtk_box_pack_start (GTK_BOX(vvbox), radio, FALSE, FALSE, 0);
    gtk_widget_show (radio);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(radio), 
				  *GET_COLOR_TYPE(fx) == SCOPE_COLOR_SOLID);
    gtk_object_set_user_data (GTK_OBJECT(radio), (gpointer)SCOPE_COLOR_SOLID);
    gtk_signal_connect (GTK_OBJECT(radio), "toggled", 
			GTK_SIGNAL_FUNC(color_type_radio_toggled), (gpointer)fx);
    
    hbox = gtk_hbox_new (FALSE, 4);
    gtk_container_add (GTK_CONTAINER(radio), hbox);
    gtk_widget_show (hbox);
    
    label = gtk_label_new ("Solid Color");
    gtk_box_pack_start (GTK_BOX(hbox), label, FALSE, FALSE, 4);
    gtk_widget_show (label);

    button = color_button_new ();
    color_button_set_color (COLOR_BUTTON(button), *GET_COLOR(fx));
    gtk_signal_connect (GTK_OBJECT(button), "color-changed", 
			GTK_SIGNAL_FUNC(color_changed), GET_COLOR(fx));
    gtk_box_pack_start (GTK_BOX(hbox), button, FALSE, FALSE, 0);
    gtk_widget_show (button);
    
    radio = gtk_radio_button_new_with_label 
      (gtk_radio_button_group(GTK_RADIO_BUTTON(radio)), "Random Colours");
    gtk_box_pack_start (GTK_BOX(vvbox), radio, FALSE, FALSE, 0);
    gtk_widget_show (radio);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(radio), 
				  *GET_COLOR_TYPE(fx) == SCOPE_COLOR_RANDOM);
    gtk_object_set_user_data (GTK_OBJECT(radio), (gpointer)SCOPE_COLOR_RANDOM);
    gtk_signal_connect (GTK_OBJECT(radio), "toggled", 
			GTK_SIGNAL_FUNC(color_type_radio_toggled), (gpointer)fx);
  }  

  return vbox;
}
