/* uniblitter.c -- a customizable blitter
 *
 * Copyright (C) 2001 Janusz Gregorczyk <jgregor@kki.net.pl>
 *
 * This file is part of xvs.
 *
 * Portions of this file are derived from Jakdaw's XMMS Plugin:
 *   Jakdaw's XMMS Plugin
 *   Copyright (c) 1999, 2000 Christopher Wilson <jakdaw@usa.net>
 *   http://www.jakdaw.ucam.org/xmms/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */
 
#include <math.h>
#include <pthread.h>
#include <string.h>

#include "misc.h"
#include "../calc/calc.h"
#include "../fx.h"
#include "../image.h"

static char *get_desc (void);
static void init (XVS_Effect *fx);
static int render (XVS_Effect *fx, XVS_Image *img, XVS_Image *img_aux,
		   gboolean beat);
static void cleanup (XVS_Effect *fx);
static GtkWidget* configure (XVS_Effect *fx);

/* Module options. */
static const XVS_Option_Desc FXUniBlitter_Option_Desc[] = {
  { "bilinear_filtering", XVS_OPT_BOOLEAN, { .bval = TRUE } },
  { "polar", XVS_OPT_BOOLEAN, { .bval = TRUE } },
  { "formula", XVS_OPT_STRING, { .sval = NULL } },
  { 0 }
};

#define GET_BILINEAR_FILTERING(fx)	(&fx->config[0].value.bval)
#define GET_POLAR(fx) 		(&fx->config[1].value.bval)
#define GET_FORMULA(fx) 	(&fx->config[2].value.sval)

/* Module data. */
typedef struct {
  pthread_mutex_t mutex;
  int w, h;			/* Previous width and height. */
  guint32 *table;
  gboolean reset;
} FXUniBlitter_Data;

/* Module description. */
XVS_Effect_Desc FXUniBlitter = {
  "Trans/UniBlitter",
  FXUniBlitter_Option_Desc,
  0,

  &get_desc,
  &init,
  &render,
  &cleanup,
  &configure
};

/* Module code. */
static char *get_desc (void) {
  return "UniBlitter";
}

static void init (XVS_Effect *fx) {
  fx->data = g_new (FXUniBlitter_Data, 1);
  
  {
    FXUniBlitter_Data *data = (FXUniBlitter_Data *)fx->data;
    
    pthread_mutex_init (&data->mutex, NULL);
    data->table = NULL;
    data->w = data->h = -1;
    data->reset = TRUE;
  }
}

static guint32 f_transform (XVS_Image *img, int x, int y, 
			    expression_t *expr, symbol_dict_t *dict,
			    double *xf, double *yf) {
  int xn, yn;

  g_assert (expr);
  g_assert (dict);

  /* Point (xf, yf) must lie within [-1..1]x[-1..1] square. */
  *xf = 2.0 * x / (img->w - 1) - 1.0;
  *yf = 2.0 * y / (img->h - 1) - 1.0;
  
  /* Transform. */
  expr_execute (expr, dict);

  /* Convert back to physical coordinates. */
  xn = (int)(((*xf + 1.0) * (img->w - 1) / 2) + 0.5);
  yn = (int)(((*yf + 1.0) * (img->h - 1) / 2) + 0.5);

  if (xn < 0 || xn >= img->w || yn < 0 || yn >= img->h) { 
    xn = x; yn = y; 
  }

  return yn * img->w + xn;
}

static guint32 f_transform_polar (XVS_Image *img, int x, int y, 
				  expression_t *expr, symbol_dict_t *dict,
				  double *r, double *fi) {
  int xn, yn;
  double xf, yf;

  g_assert (expr);
  g_assert (dict);

  /* Point (xf, yf) must lie within [-1..1]x[-1..1] square. */
  xf = 2.0 * x / (img->w - 1) - 1.0;
  yf = 2.0 * y / (img->h - 1) - 1.0;

  /* Convert to polar coordinates. */
  *r = hypot (xf, yf); 
  *fi = atan2 (yf, xf); 

  /* Transform. */
  expr_execute (expr, dict);

  /* Convert back to rectangular coordinates. */
  xf = (*r) * cos ((*fi));
  yf = (*r) * sin ((*fi));

  /* Convert back to physical coordinates. */
  xn = (int)(((xf + 1.0) * (img->w - 1) / 2) + 0.5);
  yn = (int)(((yf + 1.0) * (img->h - 1) / 2) + 0.5);

  if (xn < 0 || xn >= img->w || yn < 0 || yn >= img->h) { 
    xn = x; yn = y; 
  }

  return yn * img->w + xn;
}

static void table_init (FXUniBlitter_Data* data, XVS_Image *img, 
			gboolean filter, gboolean polar, 
			const gchar* formula) {
  int x, y;
  guint32 *tptr;
  symbol_dict_t *dict;
  expression_t *expr;
  guint32 (*transform_func) (XVS_Image *img, int x, int y,
			     expression_t *expr, symbol_dict_t *dict,
			     double *v1, double *v2);
  double *v1, *v2;

  if (!formula) return;

  /* Compile the expression. */
  dict = dict_new ();
  if (!dict) return;

  expr = expr_compile_string (formula, dict);
  if (!expr) {
    dict_free (dict);
    return;
  }

  if (polar) {
    transform_func = f_transform_polar;
    v1 = dict_variable (dict, "r");
    v2 = dict_variable (dict, "fi");
  } else {
    transform_func = f_transform;
    v1 = dict_variable (dict, "x");
    v2 = dict_variable (dict, "y");
  }

  if (filter == FALSE) { /* Without filtering */
    data->table = (guint32 *)g_malloc (img->w * img->h * sizeof(guint32));

    tptr = data->table;
    for(y = 0; y < img->h; y++) {
      for(x = 0; x < img->w; x++) {
	*tptr++ = (*transform_func) (img, x, y, expr, dict, v1, v2);
      }
    }
  } else { /* With filtering */
    data->table = (guint32 *)g_malloc (4 * img->w * img->h * sizeof(guint32));

    tptr = data->table;
    for(y = 0; y < img->h; y++) {
      for(x = 0; x < img->w; x++) {
	  *tptr++ = (*transform_func) (img, x, y, expr, dict, v1, v2);
	  *tptr++ = (*transform_func) (img, (x + 1 < img->w) ? (x + 1) : x, y,
				       expr, dict, v1, v2);
	  *tptr++ = (*transform_func) (img, x, (y + 1 < img->h) ? (y + 1) : y,
				       expr, dict, v1, v2);
	  *tptr++ = (*transform_func) (img, (x + 1 < img->w) ? (x + 1) : x,
				       (y + 1 < img->h) ? (y + 1) : y,
				       expr, dict, v1, v2);
      }
    }
  }

  dict_free (dict);
  expr_free (expr);
}

static void table_kill (FXUniBlitter_Data *data) {
  g_free (data->table);
  data->table = NULL;
}

static int render (XVS_Effect *fx, XVS_Image *img, XVS_Image *img_aux,
		   gboolean beat) {
  FXUniBlitter_Data *data = (FXUniBlitter_Data *)fx->data;

  pthread_mutex_lock (&data->mutex);

  /* Detect size change. */
  if (img->w != data->w || img->h != data->h) {
    data->w = img->w;
    data->h = img->h;
    data->reset = TRUE;
  }

  /* Reset on first initialization or option change. */
  if (data->reset) {
    table_kill (data);
    table_init (data, img, 
		*GET_BILINEAR_FILTERING(fx), *GET_POLAR(fx), *GET_FORMULA(fx));

    data->reset = FALSE;
  }

  if (!data->table) {
    pthread_mutex_unlock (&data->mutex);
    return FB_ERROR;
  }

  /* Main rendering loop. */
  if (*GET_BILINEAR_FILTERING(fx) == FALSE) { /* Without filtering */
    table_blit (img_aux->pixels, img->pixels, 
		data->table, img->w * img->h);
  } else { /* With filtering */
    table_blit_bf (img_aux->pixels, img->pixels, 
		   data->table, img->w * img->h);
  }

  pthread_mutex_unlock (&data->mutex);
      
  return FB_AUX;
}

static void cleanup (XVS_Effect *fx) {
  FXUniBlitter_Data *data = (FXUniBlitter_Data *)fx->data;

  pthread_mutex_destroy (&data->mutex);
  table_kill (data); 
 
  g_free (data);
}

static void on_formula_change (GtkEditable *editable, XVS_Effect *fx) {
  FXUniBlitter_Data *data = (FXUniBlitter_Data *)fx->data;

  pthread_mutex_lock (&data->mutex);

  g_free (*GET_FORMULA(fx));
  *GET_FORMULA(fx) = gtk_editable_get_chars (editable, 0, -1);
  data->reset = TRUE;

  pthread_mutex_unlock (&data->mutex);
}

static void filtering_toggled (GtkToggleButton *button, XVS_Effect *fx) {
  FXUniBlitter_Data *data = (FXUniBlitter_Data *)fx->data;

  pthread_mutex_lock (&data->mutex);

  *GET_BILINEAR_FILTERING(fx) = 
    gtk_toggle_button_get_active (button);
  data->reset = TRUE;

  pthread_mutex_unlock (&data->mutex);
}

static void polar_toggled (GtkToggleButton *button, XVS_Effect *fx) {
  FXUniBlitter_Data *data = (FXUniBlitter_Data *)fx->data;

  pthread_mutex_lock (&data->mutex);

  *GET_POLAR(fx) = 
    gtk_toggle_button_get_active (button);
  data->reset = TRUE;

  pthread_mutex_unlock (&data->mutex);
}

static GtkWidget* configure (XVS_Effect *fx) {
  GtkWidget *vbox, *text, *button;

  vbox = gtk_vbox_new (FALSE, 5);

  {
    button = gtk_check_button_new_with_label ("Polar coordinates");
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), 
				  *GET_POLAR(fx));
    gtk_signal_connect (GTK_OBJECT(button), "toggled", 
			GTK_SIGNAL_FUNC(polar_toggled), (gpointer)fx);
    gtk_box_pack_start (GTK_BOX(vbox), button, TRUE, TRUE, 4);
    gtk_widget_show (button);
  }

  {
    text = gtk_text_new (NULL, NULL); 

    gtk_text_freeze (GTK_TEXT(text));
    gtk_text_insert (GTK_TEXT(text), NULL, NULL, NULL,
		     *GET_FORMULA(fx), -1);
    gtk_text_thaw (GTK_TEXT(text));

    gtk_text_set_editable (GTK_TEXT(text), TRUE);
    gtk_text_set_word_wrap (GTK_TEXT(text), TRUE);

    gtk_signal_connect (GTK_OBJECT(text), "changed", 
			GTK_SIGNAL_FUNC(on_formula_change), (gpointer)fx);
    gtk_box_pack_start (GTK_BOX(vbox), text, TRUE, TRUE, 4);
    gtk_widget_show (text);
  }

  {
    button = gtk_check_button_new_with_label ("Bilinear filtering");
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), 
				  *GET_BILINEAR_FILTERING(fx));
    gtk_signal_connect (GTK_OBJECT(button), "toggled", 
			GTK_SIGNAL_FUNC(filtering_toggled), (gpointer)fx);
    gtk_box_pack_start (GTK_BOX(vbox), button, TRUE, TRUE, 4);
    gtk_widget_show (button);
  }

  return vbox;
}
