/* uniscope.c -- a customizable oscilloscope plotter
 *
 * Copyright (C) 2001 Janusz Gregorczyk <jgregor@kki.net.pl>
 *
 * This file is part of xvs.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#include <stdlib.h>
#include <glib.h>
#include <math.h>
#include <pthread.h>

#include "misc.h"
#include "../calc/calc.h"
#include "../image.h"
#include "../fx.h"
#include "../widgets/color_button.h"

static char *get_desc (void);
static void init (XVS_Effect *fx);
static int render (XVS_Effect *fx, XVS_Image *img, XVS_Image *img_aux,
		   gboolean beat);
static void cleanup (XVS_Effect *fx);
static GtkWidget* configure (XVS_Effect *fx);

enum { UNISCOPE_COLOR_SOLID = 0, UNISCOPE_COLOR_RANDOM = 1 };
enum { POINTS = 0, LINES = 1 };

/* Module options. */
static const XVS_Option_Desc FXUniScope_Option_Desc[] = {
  { "color_type", XVS_OPT_INT, { .ival = UNISCOPE_COLOR_SOLID } },
  { "color", XVS_OPT_INT, { .ival = 0x00ffffff } },
  { "mode", XVS_OPT_INT, { .ival = LINES } }, 
  { "on_init", XVS_OPT_STRING, { .sval = NULL } },
  { "per_frame", XVS_OPT_STRING, { .sval = NULL } },
  { "per_sample", XVS_OPT_STRING, { .sval = NULL } },
  { 0 }
};

#define GET_COLOR_TYPE(fx)	(&fx->config[0].value.ival)
#define GET_COLOR(fx) 		(&fx->config[1].value.ival)
#define GET_MODE(fx) 		(&fx->config[2].value.ival)
#define GET_ON_INIT(fx) 	(&fx->config[3].value.sval)
#define GET_PER_FRAME(fx) 	(&fx->config[4].value.sval)
#define GET_PER_SAMPLE(fx) 	(&fx->config[5].value.sval)

/* Module data. */
typedef struct {
  pthread_mutex_t mutex;
  expression_t *expr_on_init, *expr_per_frame, *expr_per_sample;
  symbol_dict_t *dict;
  gboolean reset;
} FXUniScope_Data;

/* Module description. */
XVS_Effect_Desc FXUniScope = {
  "Render/UniScope",
  FXUniScope_Option_Desc,
  0,
  
  &get_desc,
  &init,
  &render,
  &cleanup,
  &configure
};

/* Module code. */
static char *get_desc (void) {
  return "UniScope";
}

static void init (XVS_Effect *fx) {
  fx->data = g_new (FXUniScope_Data, 1);
  
  {
    FXUniScope_Data *data = (FXUniScope_Data *)fx->data;
    
    pthread_mutex_init (&data->mutex, NULL);

    data->expr_on_init = NULL;
    data->expr_per_frame = NULL;
    data->expr_per_sample = NULL;
    data->dict = NULL;
    data->reset = TRUE;
  }
}

static int render (XVS_Effect *fx, XVS_Image *img, XVS_Image *img_aux,
		   gboolean beat) {
  FXUniScope_Data *data = (FXUniScope_Data *)fx->data;
  guint32 color;
  int i, oldx, oldy;
  double *xf, *yf, *index, *value, *fbeat;

  pthread_mutex_lock (&data->mutex);

  /* Reset if necessary. */
  if (data->reset) {
    dict_free (data->dict);
    data->dict = dict_new ();

  /* Compile the expressions. */
    if (*GET_ON_INIT(fx) != NULL)
      data->expr_on_init = expr_compile_string
	(*GET_ON_INIT(fx), data->dict);

    if (*GET_PER_FRAME(fx) != NULL)
      data->expr_per_frame = expr_compile_string 
	(*GET_PER_FRAME(fx), data->dict);

    if (*GET_PER_SAMPLE(fx) != NULL)
      data->expr_per_sample = expr_compile_string
	(*GET_PER_SAMPLE(fx), data->dict);

    if (data->expr_on_init != NULL)
      expr_execute (data->expr_on_init, data->dict);

    data->reset = FALSE;
  }

  if (!data->dict) {
    pthread_mutex_unlock (&data->mutex);
    return FB_ERROR;
  }

  /* Define some variables. */
  xf = dict_variable (data->dict, "x");
  yf = dict_variable (data->dict, "y");
  index = dict_variable (data->dict, "index");
  value = dict_variable (data->dict, "value");
  fbeat = dict_variable (data->dict, "beat");

  *fbeat = beat ? 1 : 0;

  if (data->expr_per_frame != NULL)
    expr_execute (data->expr_per_frame, data->dict);

  if (data->expr_per_sample != NULL) {
    /* Decide on which color to use. */
    switch (*GET_COLOR_TYPE(fx)) {
    case UNISCOPE_COLOR_SOLID: color = *GET_COLOR(fx); break;
    case UNISCOPE_COLOR_RANDOM: color = random (); break;
    }
    
    /* Draw. */
    for (i = 0; i < 513; i++) {
      int x, y;

      *value = 1.0 * XVS_pcm_data[0][i & 511] / 32768.0;
      *index = i / 512.0;
      
      /* Find the coordinates. */
      expr_execute (data->expr_per_sample, data->dict);
      
      /* Convert back to physical coordinates. */
      x = (int)(((*xf + 1.0) * (img->w - 1) / 2) + 0.5);
      y = (int)(((*yf + 1.0) * (img->h - 1) / 2) + 0.5);
     
      switch (*GET_MODE(fx)) {
      case POINTS:
	if (x >= 0 && x < img->w && y >= 0 && y <img->h)
	  img->pixels[y * img->w + x] = color;
	break;
      case LINES:
	if (i != 0)
	  draw_line (img, oldx, oldy, x, y, color);
	oldx = x;
	oldy = y;
	break;
      } 
    }
  }
  
  pthread_mutex_unlock (&data->mutex);

  return FB_SINGLE;
}

static void cleanup (XVS_Effect *fx) {
  FXUniScope_Data *data = (FXUniScope_Data *)fx->data;

  pthread_mutex_destroy (&data->mutex);
 
  expr_free (data->expr_on_init);
  expr_free (data->expr_per_frame);
  expr_free (data->expr_per_sample);
  dict_free (data->dict);
  
  g_free (data);
}

static void on_init_change (GtkEditable *editable, XVS_Effect *fx) {
  FXUniScope_Data *data = (FXUniScope_Data *)fx->data;

  pthread_mutex_lock (&data->mutex);

  g_free (*GET_ON_INIT(fx));
  *GET_ON_INIT(fx) = gtk_editable_get_chars (editable, 0, -1);
  data->reset = TRUE;

  pthread_mutex_unlock (&data->mutex);
}

static void per_frame_change (GtkEditable *editable, XVS_Effect *fx) {
  FXUniScope_Data *data = (FXUniScope_Data *)fx->data;

  pthread_mutex_lock (&data->mutex);

  g_free (*GET_PER_FRAME(fx));
  *GET_PER_FRAME(fx) = gtk_editable_get_chars (editable, 0, -1);
  data->reset = TRUE;

  pthread_mutex_unlock (&data->mutex);
}

static void on_sample_change (GtkEditable *editable, XVS_Effect *fx) {
  FXUniScope_Data *data = (FXUniScope_Data *)fx->data;

  pthread_mutex_lock (&data->mutex);

  g_free (*GET_PER_SAMPLE(fx));
  *GET_PER_SAMPLE(fx) = gtk_editable_get_chars (editable, 0, -1);
  data->reset = TRUE;

  pthread_mutex_unlock (&data->mutex);
}

static void plotter_radio_toggled (GtkToggleButton *b, XVS_Effect *fx) {
  if (gtk_toggle_button_get_active (b))
    *GET_MODE(fx) = (int)gtk_object_get_user_data (GTK_OBJECT(b));
}

static void color_type_radio_toggled (GtkToggleButton *b, XVS_Effect *fx) {
  if (gtk_toggle_button_get_active (b))
    *GET_COLOR_TYPE(fx) = (int)gtk_object_get_user_data (GTK_OBJECT(b));
}

static void color_changed (Color_Button *cb, gint *value) {
  *value = cb->color;
}    

static GtkWidget* configure (XVS_Effect *fx) {
  GtkWidget *vbox, *text, *label, *frame, *hbox, *vvbox, *radio, *button;

  vbox = gtk_vbox_new (FALSE, 5);

  /* Plotter mode */  
  frame = gtk_frame_new ("Plotter mode");
  gtk_box_pack_start (GTK_BOX(vbox), frame, FALSE, FALSE, 0);
  gtk_widget_show (frame);
  
  vvbox = gtk_vbox_new (FALSE, 5);
  gtk_container_add (GTK_CONTAINER(frame), vvbox);
  gtk_widget_show (vvbox);

  {
    hbox = gtk_hbox_new (FALSE, 4);
    gtk_widget_show (hbox);
    gtk_container_add (GTK_CONTAINER(vvbox), hbox);

    radio = gtk_radio_button_new_with_label (NULL, "Points");
    gtk_box_pack_start (GTK_BOX(hbox), radio, FALSE, FALSE, 4);
    gtk_widget_show (radio);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(radio), 
				  *GET_MODE(fx) == POINTS);
    gtk_object_set_user_data (GTK_OBJECT(radio), (gpointer)POINTS);
    gtk_signal_connect (GTK_OBJECT(radio), "toggled", 
			GTK_SIGNAL_FUNC(plotter_radio_toggled), (gpointer)fx);
  
    radio = gtk_radio_button_new_with_label 
      (gtk_radio_button_group(GTK_RADIO_BUTTON(radio)), "Lines");
    gtk_box_pack_start (GTK_BOX(hbox), radio, FALSE, FALSE, 4);
    gtk_widget_show (radio);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(radio), 
				  *GET_MODE(fx) == LINES);
    gtk_object_set_user_data (GTK_OBJECT(radio), (gpointer)LINES);
    gtk_signal_connect (GTK_OBJECT(radio), "toggled", 
			GTK_SIGNAL_FUNC(plotter_radio_toggled), (gpointer)fx);
  }

  /* Colour selection */
  {
    frame = gtk_frame_new ("Colour");
    gtk_box_pack_start (GTK_BOX(vbox), frame, FALSE, FALSE, 0);
    gtk_widget_show (frame);

    vvbox = gtk_vbox_new (FALSE, 5);
    gtk_container_border_width (GTK_CONTAINER(vvbox), 4);
    gtk_container_add (GTK_CONTAINER(frame), vvbox);
    gtk_widget_show (vvbox);
    
    radio = gtk_radio_button_new (NULL);
    gtk_box_pack_start (GTK_BOX(vvbox), radio, FALSE, FALSE, 0);
    gtk_widget_show (radio);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(radio), 
				  *GET_COLOR_TYPE(fx) == UNISCOPE_COLOR_SOLID);
    gtk_object_set_user_data (GTK_OBJECT(radio), (gpointer)UNISCOPE_COLOR_SOLID);
    gtk_signal_connect (GTK_OBJECT(radio), "toggled", 
			GTK_SIGNAL_FUNC(color_type_radio_toggled), (gpointer)fx);
    
    hbox = gtk_hbox_new (FALSE, 4);
    gtk_container_add (GTK_CONTAINER(radio), hbox);
    gtk_widget_show (hbox);
    
    label = gtk_label_new ("Solid Color");
    gtk_box_pack_start (GTK_BOX(hbox), label, FALSE, FALSE, 4);
    gtk_widget_show (label);

    button = color_button_new ();
    color_button_set_color (COLOR_BUTTON(button), *GET_COLOR(fx));
    gtk_signal_connect (GTK_OBJECT(button), "color-changed", 
			GTK_SIGNAL_FUNC(color_changed), GET_COLOR(fx));
    gtk_box_pack_start (GTK_BOX(hbox), button, FALSE, FALSE, 0);
    gtk_widget_show (button);
    
    radio = gtk_radio_button_new_with_label 
      (gtk_radio_button_group(GTK_RADIO_BUTTON(radio)), "Random Colours");
    gtk_box_pack_start (GTK_BOX(vvbox), radio, FALSE, FALSE, 0);
    gtk_widget_show (radio);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(radio), 
				  *GET_COLOR_TYPE(fx) == UNISCOPE_COLOR_RANDOM);
    gtk_object_set_user_data (GTK_OBJECT(radio), (gpointer)UNISCOPE_COLOR_RANDOM);
    gtk_signal_connect (GTK_OBJECT(radio), "toggled", 
			GTK_SIGNAL_FUNC(color_type_radio_toggled), (gpointer)fx);
  }  

  /* On Init settings. */
  {
    label = gtk_label_new ("On Init:");
    gtk_box_pack_start_defaults (GTK_BOX(vbox), label);
    gtk_widget_show (label);

    text = gtk_text_new (NULL, NULL); 

    gtk_text_freeze (GTK_TEXT(text));
    gtk_text_insert (GTK_TEXT(text), NULL, NULL, NULL,
		     *GET_ON_INIT(fx), -1);
    gtk_text_thaw (GTK_TEXT(text));

    gtk_text_set_editable (GTK_TEXT(text), TRUE);
    gtk_text_set_word_wrap (GTK_TEXT(text), TRUE);

    gtk_signal_connect (GTK_OBJECT(text), "changed", 
			GTK_SIGNAL_FUNC(on_init_change), (gpointer)fx);
    gtk_box_pack_start_defaults (GTK_BOX(vbox), text);
    gtk_widget_show (text);
  }

  /* On Beat settings. */
  {
    label = gtk_label_new ("Per Frame:");
    gtk_box_pack_start_defaults (GTK_BOX(vbox), label);
    gtk_widget_show (label);

    text = gtk_text_new (NULL, NULL); 

    gtk_text_freeze (GTK_TEXT(text));
    gtk_text_insert (GTK_TEXT(text), NULL, NULL, NULL,
		     *GET_PER_FRAME(fx), -1);
    gtk_text_thaw (GTK_TEXT(text));

    gtk_text_set_editable (GTK_TEXT(text), TRUE);
    gtk_text_set_word_wrap (GTK_TEXT(text), TRUE);

    gtk_signal_connect (GTK_OBJECT(text), "changed", 
			GTK_SIGNAL_FUNC(per_frame_change), (gpointer)fx);
    gtk_box_pack_start_defaults (GTK_BOX(vbox), text);
    gtk_widget_show (text);
  }

  /* Per Point settings. */
  {
    label = gtk_label_new ("Per Sample:");
    gtk_box_pack_start_defaults (GTK_BOX(vbox), label);
    gtk_widget_show (label);

    text = gtk_text_new (NULL, NULL); 

    gtk_text_freeze (GTK_TEXT(text));
    gtk_text_insert (GTK_TEXT(text), NULL, NULL, NULL,
		     *GET_PER_SAMPLE(fx), -1);
    gtk_text_thaw (GTK_TEXT(text));

    gtk_text_set_editable (GTK_TEXT(text), TRUE);
    gtk_text_set_word_wrap (GTK_TEXT(text), TRUE);

    gtk_signal_connect (GTK_OBJECT(text), "changed", 
			GTK_SIGNAL_FUNC(on_sample_change), (gpointer)fx);
    gtk_box_pack_start_defaults (GTK_BOX(vbox), text);
    gtk_widget_show (text);
  }

  return vbox;
}
