/* zoomer.c -- a simple zooming blitter
 *
 * Copyright (C) 2001 Janusz Gregorczyk <jgregor@kki.net.pl>
 *
 * This file is part of xvs.
 *
 * Portions of this file are derived from Jakdaw's XMMS Plugin:
 *   Jakdaw's XMMS Plugin
 *   Copyright (c) 1999, 2000 Christopher Wilson <jakdaw@usa.net>
 *   http://www.jakdaw.ucam.org/xmms/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#include <pthread.h> 
#include <string.h>

#include "misc.h"
#include "../fx.h"
#include "../image.h"

static char *get_desc (void);
static void init (XVS_Effect *fx);
static int render (XVS_Effect *fx, XVS_Image *img, XVS_Image *img_aux,
		   gboolean beat);
static void cleanup (XVS_Effect *fx);
static GtkWidget* configure (XVS_Effect *fx);

/* Module options. */
const XVS_Option_Desc FXZoomer_Option_Desc[] = {
  { "bilinear_filtering", XVS_OPT_BOOLEAN, { .bval = TRUE } },
  { "zoom", XVS_OPT_FLOAT, { .fval = -0.05 } },
  { 0 }
};

#define GET_BILINEAR_FILTERING(fx)	(&fx->config[0].value.bval)
#define GET_ZOOM(fx) 			(&fx->config[1].value.fval)

/* Module data. */
typedef struct {
  pthread_mutex_t mutex;
  guint32 *table;
  int w, h;			/* Previous width and height. */
} FXZoomer_Data;

/* Module description. */
XVS_Effect_Desc FXZoomer = {
  "Trans/Zoomer",
  FXZoomer_Option_Desc,
  0,
  
  &get_desc,
  &init,
  &render,
  &cleanup,
  &configure
};

/* Module code. */
static char *get_desc (void) {
  return "Zoomer";
}

static void init (XVS_Effect *fx) {
  fx->data = g_new (FXZoomer_Data, 1);
  
  {
    FXZoomer_Data *data = (FXZoomer_Data *)fx->data;
    
    pthread_mutex_init (&data->mutex, NULL);
    data->table = NULL;
    data->w = data->h = -1;
  }
}

static guint32 f_zoom (XVS_Image *img, int x, int y, gfloat factor) {
  int xn, yn;

  xn = x + (int)((x - (img->w / 2)) * factor);
  yn = y + (int)((y - (img->h / 2)) * factor);

  if (xn < 0 || xn >= img->w || yn < 0 || yn >= img->h) { 
    xn = x; 
    yn = y; 
  }

  return yn * img->w + xn;
}

static void table_init (FXZoomer_Data* data, XVS_Image *img, 
			gboolean filter, gfloat zoom) {
  int x, y;
  guint32 *tptr;

  if (filter == FALSE) { /* Without filtering */
    data->table = (guint32 *)g_malloc (img->w * img->h * sizeof(guint32));

    tptr = data->table;
    for(y = 0; y < img->h; y++) {
      for(x = 0; x < img->w; x++) {
	*tptr++ = f_zoom (img, x, y, zoom);
      }
    }
  } else { /* With filtering */
    data->table = (guint32 *)g_malloc (4 * img->w * img->h * sizeof(guint32));

    tptr = data->table;
    for(y = 0; y < img->h; y++) {
      for(x = 0; x < img->w; x++) {
	*tptr++ = f_zoom (img, x, y, zoom);
	*tptr++ = f_zoom (img, (x + 1 < img->w) ? (x + 1) : x, y, zoom);
	*tptr++ = f_zoom (img, x, (y + 1 < img->h) ? (y + 1) : y, zoom);
	*tptr++ = f_zoom (img, (x + 1 < img->w) ? (x + 1) : x, 
			  (y + 1 < img->h) ? (y + 1) : y, zoom);
      }
    }
  }

  data->w = img->w;
  data->h = img->h;
}

static void table_kill (FXZoomer_Data *data) {
  g_free (data->table);
  data->table = NULL;
  data->w = data->h = -1;
}

static int render (XVS_Effect *fx, XVS_Image *img, XVS_Image *img_aux,
		   gboolean beat) {
  FXZoomer_Data *data = (FXZoomer_Data *)fx->data;
		   
  pthread_mutex_lock (&data->mutex);

  /* Detect first initialization or option change. */
  if (img->w != data->w || img->h != data->h || data->table == NULL) {
    table_kill (data);
    table_init (data, img, *GET_BILINEAR_FILTERING(fx), *GET_ZOOM(fx));
  }

  /* Main rendering loop. */
  if (*GET_BILINEAR_FILTERING(fx) == FALSE) { /* Without filtering */
    table_blit (img_aux->pixels, img->pixels, 
		data->table, img->w * img->h);
  } else { /* With filtering */
    table_blit_bf (img_aux->pixels, img->pixels, 
		   data->table, img->w * img->h);
  }

  pthread_mutex_unlock (&data->mutex);
      
  return FB_AUX;
}

static void cleanup (XVS_Effect *fx) {
  FXZoomer_Data *data = (FXZoomer_Data *)fx->data;

  pthread_mutex_destroy (&data->mutex);
  table_kill (data);
  
  g_free (data);
}

static void zoom_value_changed (GtkWidget *widget, XVS_Effect *fx) {
  FXZoomer_Data *data = (FXZoomer_Data *)fx->data;

  pthread_mutex_lock (&data->mutex);

  *GET_ZOOM(fx) = GTK_ADJUSTMENT(widget)->value;
  table_kill (data);

  pthread_mutex_unlock (&data->mutex);
}

static void filtering_toggled (GtkToggleButton *button, XVS_Effect *fx) {
  FXZoomer_Data *data = (FXZoomer_Data *)fx->data;

  pthread_mutex_lock (&data->mutex);

  *GET_BILINEAR_FILTERING(fx) = 
    gtk_toggle_button_get_active (button);
  table_kill (data);

  pthread_mutex_unlock (&data->mutex);
}

static GtkWidget* configure (XVS_Effect *fx) {
  GtkWidget *vbox;

  vbox = gtk_vbox_new (FALSE, 5);

  {
    GtkWidget *ahscale;
    GtkObject *aadj;

    {
      GtkWidget *hbox, *label;
      hbox = gtk_hbox_new (FALSE, 5);

      label = gtk_label_new ("In");
      gtk_box_pack_start (GTK_BOX(hbox), label, TRUE, TRUE, 4);
      gtk_widget_show (label);

      label = gtk_label_new ("Zooming speed");
      gtk_box_pack_start (GTK_BOX(hbox), label, TRUE, TRUE, 4);
      gtk_widget_show (label);

      label = gtk_label_new ("Out");
      gtk_box_pack_start (GTK_BOX(hbox), label, TRUE, TRUE, 4);
      gtk_widget_show (label);

      gtk_box_pack_start (GTK_BOX(vbox), hbox, TRUE, TRUE, 4);
      gtk_widget_show (hbox);
    }
    
    aadj = gtk_adjustment_new (*GET_ZOOM(fx), -0.2, 0.2, 0.01, 0.1, 0);
    ahscale = gtk_hscale_new (GTK_ADJUSTMENT(aadj));
    gtk_scale_set_draw_value (GTK_SCALE(ahscale), FALSE);
    gtk_scale_set_digits (GTK_SCALE(ahscale), 3);
    gtk_widget_show (ahscale);
    gtk_box_pack_start (GTK_BOX(vbox), ahscale, TRUE, TRUE, 0);
    
    gtk_signal_connect (GTK_OBJECT(aadj), "value-changed",
			GTK_SIGNAL_FUNC(zoom_value_changed), (gpointer)fx);
  }

  {
    GtkWidget *button;
    button = gtk_check_button_new_with_label ("Bilinear filtering");
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), 
				  *GET_BILINEAR_FILTERING(fx));
    gtk_signal_connect (GTK_OBJECT(button), "toggled", 
			GTK_SIGNAL_FUNC(filtering_toggled), (gpointer)fx);
    gtk_box_pack_start (GTK_BOX(vbox), button, TRUE, TRUE, 4);
    gtk_widget_show (button);
  }

  return vbox;
}
