/* fxsimple.c -- operations on simple fx modules
 *
 * Copyright (C) 2001 Janusz Gregorczyk <jgregor@kki.net.pl>
 *
 * This file is part of xvs.
 *
 * Portions of this file are derived from `Paranormal' plugin:
 *   paranormal
 *   Copyright (C) 2001 Jamie Gennis <jgennis@mindspring.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#include <glib.h>
 
#include "fx.h"

XVS_Effect *fx_create (const XVS_Effect_Desc *desc) {
  XVS_Effect *fx;
  
  g_assert (desc);
  
  /* Allocate new struct. */
  fx = g_new (XVS_Effect, 1);
  
  /* Fill in class descriptor. */
  fx->desc = desc;

  fx->config = NULL;  

  if (desc->options) {
    int opts, i;
    /* Count options. */
    for (opts = 0; desc->options[opts].id; opts++);
    
    /* Allocate space for options and fill in defaults. */
    fx->config = g_new (XVS_Option, opts);
    
    for (i = 0; i < opts; i++) {
      fx->config[i].desc = &desc->options[i];
      fx->config[i].value = desc->options[i].default_value;
    }
  }

  /* Enabled is default. */
  fx->state |= XVS_FX_ENABLED;

  fx->children = NULL;
  
  /* Initialize if needed. */
  if (desc->init)
    desc->init (fx);
  
  return fx;
}

XVS_Effect *fx_clone (const XVS_Effect *fx) {
  XVS_Effect *copy;

  g_assert (fx);

  /* Allocate new struct. */
  copy = g_new (XVS_Effect, 1);
  
  /* Fill in class descriptor, copy flags and such. */
  copy->desc = fx->desc;
  copy->state = fx->state;
  copy->children = NULL;
  copy->data = NULL;

  /* Clone options. */
  copy->config = NULL;  

  if (fx->desc->options) {
    int opts, i;
    /* Count options. */
    for (opts = 0; fx->desc->options[opts].id; opts++);
    
    /* Allocate space for options and copy iteratively. */
    copy->config = g_new (XVS_Option, opts);
    
    for (i = 0; i < opts; i++) {
      switch (fx->desc->options[i].type) {
      case XVS_OPT_INT:
      case XVS_OPT_FLOAT:
      case XVS_OPT_BOOLEAN:	/* Make a bit-copy of simple options. */
	copy->config[i] = fx->config[i];
	break;

      case XVS_OPT_STRING:	/* Duplicate strings. */
	copy->config[i].desc = fx->config[i].desc;
	copy->config[i].value = fx->config[i].value;

	if (fx->config[i].value.sval !=
	    fx->desc->options[i].default_value.sval) {
	  copy->config[i].value.sval = g_strdup (fx->config[i].value.sval);
	}
	break;
      }
    }
  }

  /* Containers need special handling. */
  if (fx->desc->flags & XVS_FLAG_CONTAINER)
    container_clone (copy, fx);

  /* Initialize if needed. */
  if (copy->desc->init)
    copy->desc->init (copy);
  
  return copy;
}

void fx_destroy (XVS_Effect *fx) {
  int i;
  
  g_assert (fx);

  /* Containers need special handling. */
  if (fx->desc->flags & XVS_FLAG_CONTAINER)
    container_destroy (fx);

  if (fx->desc->cleanup)
    fx->desc->cleanup (fx);

  /* Free memory used by options. */
  if (fx->desc->options) {
    for (i = 0; fx->desc->options[i].id; i++) {
      switch (fx->desc->options[i].type) {
      case XVS_OPT_INT:
      case XVS_OPT_FLOAT:
      case XVS_OPT_BOOLEAN:
	break;
      case XVS_OPT_STRING:
	/* We need to destroy strings created by the user. */
	if (fx->config[i].value.sval !=
	    fx->desc->options[i].default_value.sval)
	  g_free (fx->config[i].value.sval);
	break;
      }
    }
  }

  g_free (fx->config);
  
  g_free (fx);
}

void fx_render (XVS_Effect *fx,
		XVS_Image *img, XVS_Image *img_aux, gboolean beat) {
  g_assert (fx);

  if (fx->state & XVS_FX_ENABLED) {
    if (fx->desc->render)
      if (fx->desc->render (fx, img, img_aux, beat) == FB_AUX) {
	/* Swap the buffers - avoid unnecessary copying pixel data. */
	XVS_Image tmp = *img;
	*img = *img_aux;
	*img_aux = tmp;
      }
  }
}
