/* plugin.c -- main XMMS plugin module
 *
 * Copyright (C) 2001 Janusz Gregorczyk <jgregor@kki.net.pl>
 *
 * This file is part of xvs.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#include <errno.h>
#include <pthread.h>
#include <string.h>
#include <xmms/plugin.h>
#include <sys/stat.h>

#include "beat.h" 
#include "image.h" 
#include "preset.h"
#include "video/renderer.h"
#include "dialogs/about_interface.h"
#include "dialogs/config_interface.h"

static char XVS_CFG_SECTION[] = "xvs";

/* The following functions are called directly by XMMS. */
static void XVS_init (void);
static void XVS_cleanup (void);
       void XVS_about (void);
       void XVS_configure (void);
static void XVS_playback_start (void);
static void XVS_playback_stop (void);
static void XVS_render_pcm (gint16 pcm_data[2][512]);
static void XVS_render_freq (gint16 freq_data[2][256]);
 
/* This describes the plugin. */
static VisPlugin XVS_vp =
{
  NULL,		/* Filled in by xmms */
  NULL,		/* Filled in by xmms */
  0,		/* The session ID for attaching to the control socket */
  "X Visualization System",	/* The description shown in the preferences box */
  2,		/* Number of PCM channels wanted */
  2,		/* Number of freq channels wanted */
  XVS_init,	/* Called when the plugin is enabled */
  XVS_cleanup,	/* Called when the plugin is disabled */
  XVS_about,		/* Show the about box */
  XVS_configure,	/* Show the configure box */
  NULL,
  XVS_playback_start,	/* Called when playback starts */
  XVS_playback_stop,	/* Called when playback stops */
  XVS_render_pcm,	/* Render the PCM data */
  XVS_render_freq	/* Render the freq data */
};

/* Global variables. */
GtkWidget *about_window = NULL;
GtkWidget *configure_window = NULL;

static pthread_t rendering_thread;	/* Rendering thread. */
pthread_mutex_t plugin_mutex;		/* Mutex */

XVS_Effect *root_fx;
pthread_mutex_t root_fx_mutex;

static int XVS_stopped;		/* Is music stopped? */
int XVS_quit;		/* Do we need to quit?  */

gint16 XVS_pcm_data[2][512];
gint16 XVS_freq_data[2][256];

/* Function prototypes */
static void XVS_prefs_load (void);
static void XVS_prefs_save (void);
static void *XVS_rendering_thread (void *arg);

/* get_vplugin_info:
 *   This is the main entry point. XMMS calls this function to find 
 * anything else that it needs to know about this plugin.
 */
VisPlugin* get_vplugin_info (void) {
  return &XVS_vp;
}

/* Functions made available to XMMS follow. */

static void XVS_init (void) {
  XVS_quit = FALSE;
  XVS_stopped = TRUE;

  XVS_prefs_load ();

  pthread_mutex_init (&plugin_mutex, NULL);
  pthread_mutex_init (&root_fx_mutex, NULL);
  
  pthread_create (&rendering_thread, NULL, XVS_rendering_thread, NULL);
}

static void XVS_cleanup (void) {
  pthread_mutex_lock (&plugin_mutex);
  XVS_quit = TRUE; XVS_stopped = TRUE;
  pthread_mutex_unlock (&plugin_mutex);

  pthread_join (rendering_thread, NULL);

  XVS_prefs_save ();

  if (root_fx)
    fx_destroy (root_fx);
}

void XVS_about (void) {
  if (about_window == NULL) {
    about_window = create_about_window ();
    gtk_widget_show (about_window);
  }
}

void XVS_configure (void) {
  if (configure_window == NULL) {
    configure_window = create_configure_window ();
    gtk_widget_show (configure_window);
  }
}

static void XVS_playback_start (void) {
  pthread_mutex_lock (&plugin_mutex);
  XVS_stopped = FALSE;
  pthread_mutex_unlock (&plugin_mutex);
}

static void XVS_playback_stop (void) {
  pthread_mutex_lock (&plugin_mutex);
  XVS_stopped = TRUE;
  pthread_mutex_unlock (&plugin_mutex);
}

static void XVS_render_pcm (gint16 pcm_data[2][512]) {
  pthread_mutex_lock (&plugin_mutex);
  memcpy (XVS_pcm_data, pcm_data, sizeof (XVS_pcm_data));
  pthread_mutex_unlock (&plugin_mutex);
}

static void XVS_render_freq (gint16 freq_data[2][256]) {
  pthread_mutex_lock (&plugin_mutex);
  memcpy (XVS_freq_data, freq_data, sizeof (freq_data));
  pthread_mutex_unlock (&plugin_mutex);
}

/* Other functions follow. */

static gboolean check_and_create_dir (const char *dir) {
  struct stat stat_buffer;

  if (stat (dir, &stat_buffer)) {
    if (errno == ENOENT) {
      /* `dir' does not exist -- try to create it. */
      g_message ("Directory %s does not exist -- I will create it for you\n",
		 dir);
      if (mkdir(dir, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH)) {
	g_error ("Error creating dir %s: %s\n", dir, strerror (errno));
	return FALSE;
      }
    } else {
      g_error ("Couldn't stat %s: %s", dir, strerror (errno));
      return FALSE;
    }
  } else {
    /* Check if dir is really a directory. */
    if (!S_ISDIR(stat_buffer.st_mode)) {
      g_error ("%s is not a directory\n", dir);
      return FALSE;
    }
  }

  return TRUE;
}

gchar *XVS_get_preset_dir (void) {
  gchar *dirname = g_strconcat (g_get_home_dir (), "/.xmms/xvs_presets", NULL);

  check_and_create_dir (dirname);

  return dirname;
}

gchar *XVS_get_global_preset_dir (void) {
  gchar *dirname = g_strconcat (XMMS_DATA_DIR, "/xvs_presets", NULL);

  check_and_create_dir (dirname);

  return dirname;
}

static void XVS_prefs_load (void) {
  gchar *cfg_filename;
  ConfigFile *cfg;
  
  cfg_filename = g_strconcat (g_get_home_dir (), "/.xmms/config", NULL);

  /* Open config file. */
  cfg = xmms_cfg_open_file (cfg_filename);

  /* Load renderer settings. */
  renderer_cfg_load (cfg, XVS_CFG_SECTION);

  /* Load default preset. */
  {
    gchar *preset_filename;
    preset_filename = g_strconcat (XVS_get_preset_dir (), "/Default", NULL);

    pthread_mutex_lock (&root_fx_mutex);
    root_fx = preset_load (preset_filename);
    pthread_mutex_unlock (&root_fx_mutex);

    g_free (preset_filename);
  }

  xmms_cfg_free (cfg);
  g_free (cfg_filename);
}

static void XVS_prefs_save (void) {
  gchar *cfg_filename;
  ConfigFile *cfg;

  cfg_filename = g_strconcat (g_get_home_dir (), "/.xmms/config", NULL);

  /* Open config file. */
  cfg = xmms_cfg_open_file (cfg_filename);
  if (!cfg)
    cfg = xmms_cfg_new ();

  /* Save renderer settings. */
  if (cfg) {
    renderer_cfg_save (cfg, XVS_CFG_SECTION);
  }    

  /* Save current preset as default. */
  {
    gchar *preset_filename;
    preset_filename = g_strconcat (XVS_get_preset_dir (), "/Default", NULL);

    pthread_mutex_lock (&root_fx_mutex);
    if (root_fx)
      preset_save (root_fx, preset_filename);
    pthread_mutex_unlock (&root_fx_mutex);

    g_free (preset_filename);
  }
  
  xmms_cfg_write_file (cfg, cfg_filename);
  xmms_cfg_free (cfg);
  g_free (cfg_filename);
}

void XVS_disable (void) {
  XVS_vp.disable_plugin (&XVS_vp);
}

static gboolean check_finished (void) {
  gboolean retval;

  pthread_mutex_lock (&plugin_mutex);
  retval = XVS_quit;
  pthread_mutex_unlock (&plugin_mutex);

  return retval;
}

static gboolean check_stopped (void) {
  gboolean retval;

  pthread_mutex_lock (&plugin_mutex);
  retval = XVS_stopped;
  pthread_mutex_unlock (&plugin_mutex);

  return retval;
}

static void *XVS_rendering_thread (void *arg) {
  gboolean beat;

  /* Inititialize renderer. */
  renderer_init ();

  while (!check_finished ()) {
    if (renderer_handle_events ()) {
    }

    if (!check_stopped ()) {
      beat = detect_beat ();

      pthread_mutex_lock (&img_mutex);
      pthread_mutex_lock (&root_fx_mutex);
      if (root_fx)
        fx_render (root_fx, img, img_aux, beat);
      pthread_mutex_unlock (&root_fx_mutex);
      pthread_mutex_unlock (&img_mutex);
 
      renderer_display (img);
    }
  }

  /* Destroy renderer. */
  renderer_cleanup ();

  XVS_disable ();

  pthread_exit (NULL);
}
