/* preset.c -- preset i/o
 *
 * Copyright (C) 2001 Janusz Gregorczyk <jgregor@kki.net.pl>
 *
 * This file is part of xvs.
 *
 * Portions of this file are derived from `Paranormal' plugin:
 *   paranormal
 *   Copyright (C) 2001 Jamie Gennis <jgennis@mindspring.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#include <glib.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/xmlmemory.h>
#include <stdlib.h>

#include "fx.h"
#include "preset.h"

static gboolean fx_get_option (xmlNodePtr node, XVS_Option *option) {
  xmlChar *content;

  g_assert (option);

  /* Get the content. */
  content = xmlNodeGetContent (node);
  if (!content)
    return FALSE;

  switch (option->desc->type) {
  case XVS_OPT_INT:
    option->value.ival = (gint)strtol (content, NULL, 0);
    break;
  case XVS_OPT_FLOAT:
    option->value.fval = (gfloat)strtod (content, NULL);
    break;
  case XVS_OPT_STRING:
    option->value.sval = g_strdup (content);
    break;
  case XVS_OPT_BOOLEAN: {
    /* Ignore leading & trailing whitespace. */
    char *c = g_strstrip (content);

    if (g_strcasecmp (c, "TRUE") == 0)
      option->value.bval = TRUE;
    else if (g_strcasecmp (c, "FALSE") == 0)
      option->value.bval = FALSE;
  } break;
  }

  xmlFree (content);

  return TRUE;
}

static XVS_Effect *fx_parse (xmlNodePtr node) {
  const XVS_Effect_Desc *fx_driver;
  XVS_Effect *fx, *child_fx;
  int i;

  g_assert (node);

  if (xmlIsBlankNode (node) || node->type != XML_ELEMENT_NODE)
    return NULL;

  /* Unknown entity? */
  if (xmlStrcmp (node->name, "fx")) {
    g_warning ("Unknown entity: %s\n", node->name);
    return NULL;
  }

  /* Find the driver. */    
  {
    char *fx_id = xmlGetProp (node, "id");
    if (!fx_id) {
      g_warning ("\"id\" property not set\n");
      return NULL;
    }
    
    fx_driver = fx_find_driver (fx_id);
    if (!fx_driver) {
      g_warning ("Module %s not found\n", fx_id);
      return NULL;
    }
  }

  fx = fx_create (fx_driver);
  if (!fx)
    return NULL;

  if (!xmlStrcmp (xmlGetProp (node, "enabled"), "FALSE"))
     fx->state &= ~XVS_FX_ENABLED;

  /* Parse the content. */
  for (node = node->xmlChildrenNode; node; node = node->next) {
    if (xmlIsBlankNode (node) || node->type != XML_ELEMENT_NODE)
      continue;

    /* See if it's an option. */
    for (i = 0; fx->desc->options && fx->desc->options[i].id; i++) {
      if (!xmlStrcmp (node->name, 
		      (const xmlChar *)fx->desc->options[i].id))
	break;
    }

    if (fx->desc->options && fx->desc->options[i].id) {
      fx_get_option (node, &fx->config[i]);
    } else {
      /* Not an option -- consider it a child. */
      if (fx->desc->flags & XVS_FLAG_CONTAINER) {
	child_fx = fx_parse (node);
	if (child_fx)
	  container_add_fx (fx, child_fx);
      } else {
	g_warning ("Unknown entity: %s\n", node->name);
      }
    }
  }

  return fx;
}

XVS_Effect *preset_load (gchar *filename) {
  xmlDocPtr doc;
  xmlNodePtr node;
  XVS_Effect *fx;

  /* Load and parse xml document. */
  doc = xmlParseFile (filename);
  if (!doc)
    return NULL;

  node = xmlDocGetRootElement (doc);
  if (node)
    fx = fx_parse (node);

  xmlFreeDoc (doc);

  return fx;
}

static xmlNodePtr fx_save (const XVS_Effect *fx) {
  xmlNodePtr node;
  int i;

  g_assert (fx);

  /* Create a new xml node. */
  /* Simple fx's and containers are distinguished 
     by driver flags & content, not by element name. */
  node = xmlNewNode (NULL, "fx");
  xmlSetProp (node, "id", fx->desc->id);
  xmlSetProp (node, "enabled", (fx->state & XVS_FX_ENABLED) ? "TRUE" : "FALSE");

  /* Save options. */
  if (fx->desc->options) {
    for (i = 0; fx->desc->options[i].id; i++) {
      xmlNodePtr option = xmlNewChild (node, NULL, fx->config[i].desc->id, 
				       NULL);
      gchar *content = NULL;

      switch (fx->config[i].desc->type) {
      case XVS_OPT_INT:
	content = g_strdup_printf ("%d", fx->config[i].value.ival);
	break;
      case XVS_OPT_FLOAT:
	content = g_strdup_printf ("%f", fx->config[i].value.fval);
	break;
      case XVS_OPT_STRING:
	content = g_strdup (fx->config[i].value.sval);
	break;
      case XVS_OPT_BOOLEAN:
	content = g_strdup_printf ("%s", fx->config[i].value.bval ? 
				   "TRUE" : "FALSE");
	break;
      }
      
      if (content)
	xmlNodeSetContent (option, content);
      
      g_free (content);
    }
  }
  
  /* Save children. */
  if (fx->desc->flags & XVS_FLAG_CONTAINER) {
    GList *child = fx->children;
    while (child) {
      xmlAddChild (node, fx_save ((XVS_Effect *)child->data));
      child = child->next;
    }
  }

  return node;
}

gboolean preset_save (const XVS_Effect *fx, gchar* filename) {
  xmlDocPtr doc;
  xmlNodePtr node;
  
  g_assert (fx);

  /* Create a new xml document tree. */
  doc = xmlNewDoc("1.0");
  
  node = fx_save (fx);

  xmlDocSetRootElement (doc, node);

  /* Save xml tree to the file. */
  xmlSaveFormatFile (filename, doc, TRUE);

  xmlFreeDoc (doc);

  return TRUE;
}
