/* gtkrenderer.c -- a GTK/GDK renderer
 *
 * Copyright (C) 2001 Janusz Gregorczyk <jgregor@kki.net.pl>
 *
 * This file is part of xvs.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#include <gtk/gtk.h>
#include <xmms/plugin.h>

#include "../dialogs/popup_interface.h"
#include "renderer.h"
#include "gtkrenderer.h"

/* Global variables. */
static GtkWidget *window;		/* Main window. */
static GtkWidget *drawing_area;		/* Main window's drawing area. */

/* Function prototypes. */
static gint mouse_button_event (GtkWidget *widget, GdkEventButton *event);
static gint delete_event (GtkWidget *widget, GdkEvent *event, gpointer data);
static gint configure_event (GtkWidget *widget, GdkEventConfigure *event);

/* Functions made available to XMMS follow. */

void gtk_renderer_init (int width, int height) {
  /* Create the window */
  window = gtk_window_new (GTK_WINDOW_DIALOG);
  gtk_window_set_title (GTK_WINDOW(window), "XVS");

  gtk_window_set_policy (GTK_WINDOW(window), TRUE, TRUE, FALSE);
  gtk_signal_connect (GTK_OBJECT(window), "delete_event",
		      GTK_SIGNAL_FUNC(delete_event), NULL);
  gtk_signal_connect (GTK_OBJECT(window), "configure_event",
		      GTK_SIGNAL_FUNC(configure_event), NULL);
  
  /* Create drawing area inside the window. */
  drawing_area = gtk_drawing_area_new();
  gtk_drawing_area_size (GTK_DRAWING_AREA(drawing_area), width, height);
  gtk_container_add (GTK_CONTAINER(window), drawing_area);
  gtk_widget_show (drawing_area);

  /* Arrange for mouse events to be detected */
  gtk_signal_connect (GTK_OBJECT(drawing_area), "button_press_event",
		      GTK_SIGNAL_FUNC(mouse_button_event), NULL);
  gtk_widget_set_events (drawing_area, GDK_BUTTON_PRESS_MASK);

  /* Show the window. */   
  gtk_widget_show (window);
}

void gtk_renderer_cleanup (void) {
  /* Destroy window. */
  if (window) {
    gtk_widget_destroy (window);
    window = NULL;
  }
}

void gtk_renderer_display (XVS_Image *img) {
  GDK_THREADS_ENTER();

  gdk_draw_rgb_32_image (drawing_area->window, drawing_area->style->white_gc,
                         0, 0, img->w, img->h,
		         GDK_RGB_DITHER_NONE, (guchar*)img->pixels, img->w << 2);

  GDK_THREADS_LEAVE();
}

void gtk_renderer_resize (int width, int height) {
  GDK_THREADS_ENTER();
  gtk_drawing_area_size (GTK_DRAWING_AREA(drawing_area), width, height);
  GDK_THREADS_LEAVE();
}

/* Here go the signal handlers. */

static gint mouse_button_event (GtkWidget *widget, GdkEventButton *event) {
  GtkWidget *menu = create_popup_menu ();

  if (event->type == GDK_BUTTON_PRESS 
      && (event->button == 1 || event->button == 3)) {
    gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL, 
   	  	    event->button, event->time);
  }
  
  return TRUE;
}

static gint delete_event (GtkWidget *widget, GdkEvent *event, gpointer data) {
  renderer_request_disable ();
  return TRUE;
}

static gint configure_event (GtkWidget *widget, GdkEventConfigure *event) {
  gint width, height;

  /* Detect change in window size */
  width = event->width;
  height = event->height;

  if (width != img->w || height != img->h) {
    renderer_request_resize (width, height);
  } 

  return FALSE;
}
