/* renderer.c -- a generic renderer interface
 *
 * Copyright (C) 2001 Janusz Gregorczyk <jgregor@kki.net.pl>
 *
 * This file is part of xvs.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#include <glib.h>
#include <gtk/gtk.h>
#include <pthread.h>

#include "renderer.h"
#include "gtkrenderer.h"

XVS_Image *img;			/* Image buffer */
XVS_Image *img_aux;		/* Auxilliary image buffer */
pthread_mutex_t img_mutex;

static int renderer_width, renderer_height;

static gboolean request_resize, request_disable;
static int request_width, request_height;
static pthread_mutex_t request_mutex;

extern void XVS_disable (void);
extern int XVS_quit;
static pthread_mutex_t plugin_mutex;

/* */

void renderer_init (void) {
  request_disable = FALSE;
  request_resize = FALSE;

  img = XVS_img_create (renderer_width, renderer_height);
  img_aux = XVS_img_create (renderer_width, renderer_height);
  gtk_renderer_init (renderer_width, renderer_height);

  pthread_mutex_init (&img_mutex, NULL);
  pthread_mutex_init (&request_mutex, NULL);
}

void renderer_cleanup (void) {
  gtk_renderer_cleanup ();

  XVS_img_free (img); img = NULL;
  XVS_img_free (img_aux); img_aux = NULL;
}

void renderer_display (XVS_Image *img) {
  pthread_mutex_lock (&img_mutex);
  gtk_renderer_display (img);
  pthread_mutex_unlock (&img_mutex);
}

void renderer_cfg_load (ConfigFile *cfg, gchar *section) {
  renderer_width = renderer_height = 256;

  if (cfg) {
    xmms_cfg_read_int (cfg, section, "width", &renderer_width);
    xmms_cfg_read_int (cfg, section, "height", &renderer_height);
  }
}

void renderer_cfg_save (ConfigFile *cfg, gchar *section) {
  xmms_cfg_write_int (cfg, section, "width", renderer_width);
  xmms_cfg_write_int (cfg, section, "height", renderer_height);
}

static void renderer_resize (int width, int height) {
  pthread_mutex_lock (&img_mutex);

  renderer_width = width;
  renderer_height = height;

  img = XVS_img_resize (img, width, height);
  img_aux = XVS_img_resize (img_aux, width, height);

  pthread_mutex_unlock (&img_mutex);
}

int renderer_handle_events (void) {
  int state_change = FALSE;

  pthread_mutex_lock (&request_mutex);
  
  if (request_resize) { 
    renderer_resize (request_width, request_height);
    gtk_renderer_resize (request_width, request_height);

    request_resize = FALSE;
    state_change = TRUE;
  }

  if (request_disable) {
    pthread_mutex_lock (&plugin_mutex);
    XVS_quit = TRUE;
    pthread_mutex_unlock (&plugin_mutex);
    request_disable = FALSE;
    state_change = TRUE;
  }
  
  pthread_mutex_unlock (&request_mutex);

  return state_change;
}

void renderer_request_resize (int width, int height) {
  pthread_mutex_lock (&request_mutex);
  request_resize = TRUE;
  request_width = width;
  request_height = height;
  pthread_mutex_unlock (&request_mutex);
}

void renderer_request_disable () {
  pthread_mutex_lock (&request_mutex);
  request_disable = TRUE;
  pthread_mutex_unlock (&request_mutex);
}
