/* color_button.c -- a button with color preview
 *
 * Copyright (C) 2001 Janusz Gregorczyk <jgregor@kki.net.pl>
 *
 * This file is part of xvs.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#include <gtk/gtk.h>

#include "color_button.h"

static void color_button_class_init (Color_ButtonClass *class);
static void color_button_init (Color_Button *button);

static void color_clicked (Color_Button *button, gpointer user_data);
static void color_preview_set_color (GtkPreview *preview, guint32 color);

enum {
  COLOR_CHANGED_SIGNAL,
  LAST_SIGNAL
};

static gint color_button_signals[LAST_SIGNAL] = { 0 };

guint color_button_get_type (void) {
  static guint cb_type = 0;

  if (!cb_type) {
    GtkTypeInfo cb_info = {
      "Color_Button",
      sizeof (Color_Button),
      sizeof (Color_ButtonClass),
      (GtkClassInitFunc) color_button_class_init,
      (GtkObjectInitFunc) color_button_init,
      (GtkArgSetFunc) NULL,
      (GtkArgGetFunc) NULL
    };
    
    cb_type = gtk_type_unique (gtk_button_get_type (), &cb_info);
  }

  return cb_type;
}

static void color_button_class_init (Color_ButtonClass *class) {
  GtkObjectClass *object_class;

  object_class = (GtkObjectClass*) class;
  
  color_button_signals[COLOR_CHANGED_SIGNAL] = 
    gtk_signal_new ("color-changed",
		    GTK_RUN_FIRST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (Color_ButtonClass, color_changed),
		    gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);

  gtk_object_class_add_signals (object_class, color_button_signals, LAST_SIGNAL);

  class->color_changed = NULL;
}

static void color_button_init (Color_Button *button) {
  GtkWidget *preview;

  /* Create a preview widget and put it inside the button. */
  preview = gtk_preview_new (GTK_PREVIEW_COLOR);
  gtk_preview_size (GTK_PREVIEW(preview), 32, 16);
  gtk_container_add (GTK_CONTAINER(button), preview);
  gtk_widget_show (preview);

  color_preview_set_color (GTK_PREVIEW(preview), button->color);

  gtk_signal_connect (GTK_OBJECT(button), "clicked", 
		      GTK_SIGNAL_FUNC(color_clicked), NULL);

  button->preview = GTK_PREVIEW(preview);
}

GtkWidget* color_button_new (void) {
  return GTK_WIDGET(gtk_type_new (color_button_get_type ()));
}

/* */

void color_button_set_color (Color_Button *button, guint32 color) {
  button->color = color;
  color_preview_set_color (button->preview, color);
}

static void color_preview_set_color (GtkPreview *preview, guint32 color) {
  /* FIXME: This piece of code is particularly ugly. Yuck! */
  int i;
  guchar buf[3 * 32];

  for(i = 0; i < 32; i++) {
    buf[3 * i]     = (color & 0xff);
    buf[3 * i + 1] = (color & 0xff00) >> 8;
    buf[3 * i + 2] = (color & 0xff0000) >> 16;
  }
 
  for(i = 0; i < 16; i++)
    gtk_preview_draw_row (GTK_PREVIEW(preview), buf, 0, i, 32);
  
  gtk_widget_draw (GTK_WIDGET(preview), NULL);
}

static void csel_ok (GtkWidget *w, GtkWidget *csel) {
  Color_Button *button;
  gdouble colorf[3];
  guint32 color;

  gtk_color_selection_get_color 
    (GTK_COLOR_SELECTION(GTK_COLOR_SELECTION_DIALOG(csel)->colorsel), &colorf[0]);
 
  color = 
    (((guint32)(colorf[0] * 255.0)
   | ((guint32)(colorf[1] * 255.0) << 8))
   | ((guint32)(colorf[2] * 255.0) << 16));

  button = (Color_Button *)gtk_object_get_user_data (GTK_OBJECT(csel));

  color_button_set_color (button, color);
  gtk_signal_emit (GTK_OBJECT (button), 
		   color_button_signals[COLOR_CHANGED_SIGNAL]);

  gtk_widget_destroy (csel);
}

static gint csel_delete_event (GtkWidget *w, GdkEvent *e, GtkWidget *csel) {
  return FALSE;
}	

static void color_clicked (Color_Button *button, gpointer user_data) {
  GtkWidget *csel;
  gdouble color[3];
  
  color[0] = (button->color & 0xff) / 255.0;
  color[1] = ((button->color >> 8) & 0xff) / 255.0;
  color[2] = ((button->color >> 16) & 0xff) / 255.0;
  
  csel = gtk_color_selection_dialog_new ("Color");
  gtk_window_set_modal
    (GTK_WINDOW(&(GTK_COLOR_SELECTION_DIALOG(csel)->window)), TRUE);
  gtk_widget_hide (GTK_COLOR_SELECTION_DIALOG(csel)->help_button);
  gtk_widget_hide (GTK_COLOR_SELECTION_DIALOG(csel)->cancel_button);
  gtk_color_selection_set_color
    (GTK_COLOR_SELECTION(GTK_COLOR_SELECTION_DIALOG(csel)->colorsel), &color[0]);
  gtk_object_set_user_data (GTK_OBJECT(csel), button);
  gtk_widget_show (csel);
  
  gtk_signal_connect (GTK_OBJECT(GTK_COLOR_SELECTION_DIALOG(csel)->ok_button), 
		      "clicked", GTK_SIGNAL_FUNC(csel_ok), csel);
  gtk_signal_connect (GTK_OBJECT(csel), "delete_event", 
		      GTK_SIGNAL_FUNC(csel_delete_event), csel);
}
